#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock
from webob import exc

from manila import context
from manila.services import api as service_api
from manila import test


class ServicesApiTest(test.TestCase):

    def setUp(self):
        super(ServicesApiTest, self).setUp()
        self.context = context.get_admin_context()
        self.share_rpcapi = mock.Mock()
        self.share_rpcapi.ensure_shares = mock.Mock()
        self.services_api = service_api.API()
        self.mock_object(
            self.services_api, 'share_rpcapi', self.share_rpcapi
        )

    def test_ensure_shares(self):
        host = 'fake_host@fakebackend'
        fake_service = {
            'id': 'fake_service_id',
            'state': 'up'
        }

        self.services_api.ensure_shares(self.context, fake_service, host)

        self.share_rpcapi.ensure_driver_resources.assert_called_once_with(
            self.context, host
        )

    def test_ensure_shares_host_down(self):
        host = 'fake_host@fakebackend'
        fake_service = {
            'id': 'fake_service_id',
            'state': 'down'
        }

        self.assertRaises(
            exc.HTTPConflict,
            self.services_api.ensure_shares,
            self.context,
            fake_service,
            host
        )

        self.share_rpcapi.ensure_shares.assert_not_called()



# Copyright (C) 2022 China Telecom Digital Intelligence.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Handles all requests relating to transferring ownership of shares.
"""


import hashlib
import hmac
import os

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils
from oslo_utils import strutils

from manila.common import constants
from manila.db import base
from manila import exception
from manila.i18n import _
from manila import policy
from manila import quota
from manila.share import api as share_api
from manila.share import share_types
from manila.share import utils as share_utils


share_transfer_opts = [
    cfg.IntOpt('share_transfer_salt_length',
               default=8,
               help='The number of characters in the salt.',
               min=8,
               max=255),
    cfg.IntOpt('share_transfer_key_length',
               default=16,
               help='The number of characters in the autogenerated auth key.',
               min=16,
               max=255),
]

CONF = cfg.CONF
CONF.register_opts(share_transfer_opts)

LOG = logging.getLogger(__name__)
QUOTAS = quota.QUOTAS


class API(base.Base):
    """API for interacting share transfers."""

    def __init__(self):
        self.share_api = share_api.API()
        super().__init__()

    def get(self, context, transfer_id):
        transfer = self.db.transfer_get(context, transfer_id)
        return transfer

    def delete(self, context, transfer_id):
        """Delete a share transfer."""
        transfer = self.db.transfer_get(context, transfer_id)
        policy.check_policy(context, 'share_transfer', 'delete', target_obj={
            'project_id': transfer['source_project_id']})
        update_share_status = True
        share_ref = None
        try:
            share_ref = self.db.share_get(context, transfer.resource_id)
        except exception.NotFound:
            update_share_status = False
        if update_share_status:
            share_instance = share_ref['instance']
        if share_ref['status'] != constants.STATUS_AWAITING_TRANSFER:
            msg = (_('Transfer %(transfer_id)s: share id %(share_id)s '
                     'expected in awaiting_transfer state.'))
            msg_payload = {'transfer_id': transfer_id,
                           'share_id': share_ref['id']}
            LOG.error(msg, msg_payload)
            raise exception.InvalidShare(reason=msg)
        if update_share_status:
            share_utils.notify_about_share_usage(context, share_ref,
                                                 share_instance,
                                                 "transfer.delete.start")
        self.db.transfer_destroy(context, transfer_id,
                                 update_share_status=update_share_status)
        if update_share_status:
            share_utils.notify_about_share_usage(context, share_ref,
                                                 share_instance,
                                                 "transfer.delete.end")
        LOG.info('Transfer %s has been deleted successful.', transfer_id)

    def get_all(self, context, limit=None, sort_key=None,
                sort_dir=None, filters=None, offset=None):
        filters = filters or {}
        all_tenants = strutils.bool_from_string(filters.pop('all_tenants',
                                                            'false'))
        query_by_project = False

        if all_tenants:
            try:
                policy.check_policy(context, 'share_transfer',
                                    'get_all_tenant')
            except exception.PolicyNotAuthorized:
                query_by_project = True
        else:
            query_by_project = True

        if query_by_project:
            transfers = self.db.transfer_get_all_by_project(
                context, context.project_id,
                limit=limit, sort_key=sort_key, sort_dir=sort_dir,
                filters=filters, offset=offset)
        else:
            transfers = self.db.transfer_get_all(context,
                                                 limit=limit,
                                                 sort_key=sort_key,
                                                 sort_dir=sort_dir,
                                                 filters=filters,
                                                 offset=offset)

        return transfers

    def _get_random_string(self, length):
        """Get a random hex string of the specified length."""
        rndstr = ""

        # Note that the string returned by this function must contain only
        # characters that the recipient can enter on their keyboard. The
        # function sha256().hexdigit() achieves this by generating a hash
        # which will only contain hexadecimal digits.
        while len(rndstr) < length:
            rndstr += hashlib.sha256(os.urandom(255)).hexdigest()

        return rndstr[0:length]

    def _get_crypt_hash(self, salt, auth_key):
        """Generate a random hash based on the salt and the auth key."""
        def _format_str(input_str):
            if not isinstance(input_str, (bytes, str)):
                input_str = str(input_str)
            if isinstance(input_str, str):
                input_str = input_str.encode('utf-8')
            return input_str
        salt = _format_str(salt)
        auth_key = _format_str(auth_key)
        return hmac.new(salt, auth_key, hashlib.sha256).hexdigest()

    def create(self, context, share_id, display_name):
        """Creates an entry in the transfers table."""
        LOG.debug("Generating transfer record for share %s", share_id)
        try:
            share_ref = self.share_api.get(context, share_id)
        except exception.NotFound:
            msg = _("Share specified was not found.")
            raise exception.InvalidShare(reason=msg)
        policy.check_policy(context, "share_transfer", "create",
                            target_obj=share_ref)
        share_instance = share_ref['instance']

        mount_point_name = share_instance['mount_point_name']
        if (mount_point_name and
                mount_point_name.startswith(share_ref['project_id'])):
            msg = _('Share %s has a custom mount_point_name %s.'
                    ' This has the project_id encoded in it.'
                    ' Transferring such'
                    ' a share isn\'t supported') % (share_ref['name'],
                                                    mount_point_name)
            raise exception.Invalid(reason=msg)

        if share_ref['status'] != "available":
            raise exception.InvalidShare(reason=_("Share's status must be "
                                                  "available"))
        if share_ref['share_network_id']:
            raise exception.InvalidShare(reason=_(
                "Shares exported over share networks cannot be transferred."))
        if share_ref['share_group_id']:
            raise exception.InvalidShare(reason=_(
                "Shares within share groups cannot be transferred."))

        if share_ref.has_replicas:
            raise exception.InvalidShare(reason=_(
                "Shares with replicas cannot be transferred."))

        snapshots = self.db.share_snapshot_get_all_for_share(context, share_id)
        for snapshot in snapshots:
            if snapshot['status'] != "available":
                msg = _("Snapshot: %s status must be "
                        "available") % snapshot['id']
                raise exception.InvalidSnapshot(reason=msg)

        share_utils.notify_about_share_usage(context, share_ref,
                                             share_instance,
                                             "transfer.create.start")
        # The salt is just a short random string.
        salt = self._get_random_string(CONF.share_transfer_salt_length)
        auth_key = self._get_random_string(CONF.share_transfer_key_length)
        crypt_hash = self._get_crypt_hash(salt, auth_key)

        transfer_rec = {'resource_type': constants.SHARE_RESOURCE_TYPE,
                        'resource_id': share_id,
                        'display_name': display_name,
                        'salt': salt,
                        'crypt_hash': crypt_hash,
                        'expires_at': None,
                        'source_project_id': share_ref['project_id']}

        try:
            transfer = self.db.transfer_create(context, transfer_rec)
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.error("Failed to create transfer record for %s", share_id)
        share_utils.notify_about_share_usage(context, share_ref,
                                             share_instance,
                                             "transfer.create.end")
        return {'id': transfer['id'],
                'resource_type': transfer['resource_type'],
                'resource_id': transfer['resource_id'],
                'display_name': transfer['display_name'],
                'auth_key': auth_key,
                'created_at': transfer['created_at'],
                'source_project_id': transfer['source_project_id'],
                'destination_project_id': transfer['destination_project_id'],
                'accepted': transfer['accepted'],
                'expires_at': transfer['expires_at']}

    def _handle_snapshot_quota(self, context, snapshots, donor_id):
        snapshots_num = len(snapshots)
        share_snap_sizes = 0
        for snapshot in snapshots:
            share_snap_sizes += snapshot['size']
        try:
            reserve_opts = {'snapshots': snapshots_num,
                            'gigabytes': share_snap_sizes}
            reservations = QUOTAS.reserve(context, **reserve_opts)
        except exception.OverQuota as e:
            reservations = None
            overs = e.kwargs['overs']
            usages = e.kwargs['usages']
            quotas = e.kwargs['quotas']

            def _consumed(name):
                return (usages[name]['reserved'] + usages[name]['in_use'])

            if 'snapshot_gigabytes' in overs:
                msg = ("Quota exceeded for %(s_pid)s, tried to accept "
                       "%(s_size)sG snapshot (%(d_consumed)dG of "
                       "%(d_quota)dG already consumed).")
                LOG.warning(msg, {
                    's_pid': context.project_id,
                    's_size': share_snap_sizes,
                    'd_consumed': _consumed('snapshot_gigabytes'),
                    'd_quota': quotas['snapshot_gigabytes']})
                raise exception.SnapshotSizeExceedsAvailableQuota()
            elif 'snapshots' in overs:
                msg = ("Quota exceeded for %(s_pid)s, tried to accept "
                       "%(s_num)s snapshot (%(d_consumed)d of "
                       "%(d_quota)d already consumed).")
                LOG.warning(msg, {'s_pid': context.project_id,
                                  's_num': snapshots_num,
                                  'd_consumed': _consumed('snapshots'),
                                  'd_quota': quotas['snapshots']})
                raise exception.SnapshotLimitExceeded(
                    allowed=quotas['snapshots'])

        try:
            reserve_opts = {'snapshots': -snapshots_num,
                            'gigabytes': -share_snap_sizes}
            donor_reservations = QUOTAS.reserve(context,
                                                project_id=donor_id,
                                                **reserve_opts)
        except exception.OverQuota:
            donor_reservations = None
            LOG.exception("Failed to update share providing snapshots quota:"
                          " Over quota.")

        return reservations, donor_reservations

    @staticmethod
    def _check_share_type_access(context, share_type_id, share_id):
        share_type = share_types.get_share_type(
            context, share_type_id, expected_fields=['projects'])
        if not share_type['is_public']:
            if context.project_id not in share_type['projects']:
                msg = _("Share type of share %(share_id)s is not public, "
                        "and current project can not access the share "
                        "type ") % {'share_id': share_id}
                LOG.error(msg)
                raise exception.InvalidShare(reason=msg)

    def _check_transferred_project_quota(self, context, share_ref_size):
        try:
            reserve_opts = {'shares': 1, 'gigabytes': share_ref_size}
            reservations = QUOTAS.reserve(context,
                                          **reserve_opts)
        except exception.OverQuota as exc:
            reservations = None
            self.share_api.check_if_share_quotas_exceeded(context, exc,
                                                          share_ref_size)
        return reservations

    @staticmethod
    def _check_donor_project_quota(context, donor_id, share_ref_size,
                                   transfer_id):
        try:
            reserve_opts = {'shares': -1, 'gigabytes': -share_ref_size}
            donor_reservations = QUOTAS.reserve(context.elevated(),
                                                project_id=donor_id,
                                                **reserve_opts)
        except Exception:
            donor_reservations = None
            LOG.exception("Failed to update quota donating share"
                          " transfer id %s", transfer_id)
        return donor_reservations

    @staticmethod
    def _check_snapshot_status(snapshots, transfer_id):
        for snapshot in snapshots:
            # Only check snapshot with instances
            if snapshot.get('status'):
                if snapshot['status'] != 'available':
                    msg = (_('Transfer %(transfer_id)s: Snapshot '
                             '%(snapshot_id)s is not in the expected '
                             'available state.')
                           % {'transfer_id': transfer_id,
                              'snapshot_id': snapshot['id']})
                    LOG.error(msg)
                    raise exception.InvalidSnapshot(reason=msg)

    def accept(self, context, transfer_id, auth_key, clear_rules=False):
        """Accept a share that has been offered for transfer."""
        # We must use an elevated context to make sure we can find the
        # transfer.
        transfer = self.db.transfer_get(context.elevated(), transfer_id)

        crypt_hash = self._get_crypt_hash(transfer['salt'], auth_key)
        if crypt_hash != transfer['crypt_hash']:
            msg = (_("Attempt to transfer %s with invalid auth key.") %
                   transfer_id)
            LOG.error(msg)
            raise exception.InvalidAuthKey(reason=msg)

        share_id = transfer['resource_id']
        try:
            # We must use an elevated context to see the share that is still
            # owned by the donor.
            share_ref = self.share_api.get(context.elevated(), share_id)
        except exception.NotFound:
            msg = _("Share specified was not found.")
            raise exception.InvalidShare(reason=msg)
        share_instance = share_ref['instance']
        if share_ref['status'] != constants.STATUS_AWAITING_TRANSFER:
            msg = (_('Transfer %(transfer_id)s: share id %(share_id)s '
                     'expected in awaiting_transfer state.')
                   % {'transfer_id': transfer_id, 'share_id': share_id})
            LOG.error(msg)
            raise exception.InvalidShare(reason=msg)
        share_ref_size = share_ref['size']
        share_type_id = share_ref.get('share_type_id')
        # check share type access
        if share_type_id:
            self._check_share_type_access(context, share_type_id, share_id)

        # check per share quota limit
        self.share_api.check_is_share_size_within_per_share_quota_limit(
            context, share_ref_size)

        # check accept transferred project quotas
        reservations = self._check_transferred_project_quota(
            context, share_ref_size)

        # check donor project quotas
        donor_id = share_ref['project_id']
        donor_reservations = self._check_donor_project_quota(
            context, donor_id, share_ref_size, transfer_id)

        snap_res = None
        snap_donor_res = None
        accept_snapshots = False
        snapshots = self.db.share_snapshot_get_all_for_share(
            context.elevated(), share_id)
        if snapshots:
            self._check_snapshot_status(snapshots, transfer_id)
            accept_snapshots = True
            snap_res, snap_donor_res = self._handle_snapshot_quota(
                context, snapshots, share_ref['project_id'])

        share_utils.notify_about_share_usage(context, share_ref,
                                             share_instance,
                                             "transfer.accept.start")
        try:
            self.share_api.transfer_accept(context,
                                           share_ref,
                                           context.user_id,
                                           context.project_id,
                                           clear_rules=clear_rules)
            # Transfer ownership of the share now, must use an elevated
            # context.
            self.db.transfer_accept(context.elevated(),
                                    transfer_id,
                                    context.user_id,
                                    context.project_id,
                                    accept_snapshots=accept_snapshots)
            if reservations:
                QUOTAS.commit(context, reservations)
            if snap_res:
                QUOTAS.commit(context, snap_res)
            if donor_reservations:
                QUOTAS.commit(context, donor_reservations, project_id=donor_id)
            if snap_donor_res:
                QUOTAS.commit(context, snap_donor_res, project_id=donor_id)
            LOG.info("share %s has been transferred.", share_id)
        except Exception:
            with excutils.save_and_reraise_exception():
                try:
                    # storage try to rollback
                    self.share_api.transfer_accept(context,
                                                   share_ref,
                                                   share_ref['user_id'],
                                                   share_ref['project_id'])
                    # db try to rollback
                    self.db.transfer_accept_rollback(
                        context.elevated(), transfer_id,
                        share_ref['user_id'], share_ref['project_id'],
                        rollback_snap=accept_snapshots)
                finally:
                    if reservations:
                        QUOTAS.rollback(context, reservations)
                    if snap_res:
                        QUOTAS.rollback(context, snap_res)
                    if donor_reservations:
                        QUOTAS.rollback(context, donor_reservations,
                                        project_id=donor_id)
                    if snap_donor_res:
                        QUOTAS.rollback(context, snap_donor_res,
                                        project_id=donor_id)

        share_utils.notify_about_share_usage(context, share_ref,
                                             share_instance,
                                             "transfer.accept.end")



# Copyright (c) 2023 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
import os

import requests_mock

from manila.share.drivers.dell_emc.plugins.powerstore import client
from manila import test


class TestClient(test.TestCase):

    REST_IP = "192.168.0.110"
    NAS_SERVER_NAME = "powerstore-nasserver"
    NAS_SERVER_ID = "6423d56e-eaf3-7424-be0b-1a9efb93188b"
    NAS_SERVER_IP = "192.168.11.23"
    NFS_EXPORT_NAME = "powerstore-nfs-share"
    NFS_EXPORT_SIZE = 3221225472
    NFS_EXPORT_NEW_SIZE = 6221225472
    FILESYSTEM_ID = "6454e9a9-a698-e9bc-ca61-1a9efb93188b"
    NFS_EXPORT_ID = "6454ec18-7b8d-1532-1b8a-1a9efb93188b"
    RW_HOSTS = "192.168.1.10"
    RO_HOSTS = "192.168.1.11"
    SMB_SHARE_NAME = "powerstore-smb-share"
    SMB_SHARE_ID = "64927ae9-3403-6930-a784-f227b9987c54"
    RW_USERS = "user1"
    RO_USERS = "user2"
    SNAPSHOT_NAME = "powerstore-nfs-share-snap"
    SNAPSHOT_ID = "6454ea29-09c3-030e-cfc3-1a9efb93188b"
    CLONE_ID = "64560f05-e677-ec2a-7fcf-1a9efb93188b"
    CLONE_NAME = "powerstore-nfs-share-snap-clone"
    CLUSTER_ID = "0"

    CLIENT_OPTIONS = {
        "rest_ip": REST_IP,
        "rest_username": "admin",
        "rest_password": "pwd",
        "verify_certificate": False,
        "certificate_path": None
    }

    def setUp(self):
        super(TestClient, self).setUp()

        self._mock_url = "https://%s/api/rest" % self.REST_IP
        self.client = client.PowerStoreClient(**self.CLIENT_OPTIONS)
        self.mockup_file_base = os.path.join(
            os.path.dirname(os.path.realpath(__file__)), 'mockup')

    def _getJsonFile(self, filename):
        f = open(os.path.join(self.mockup_file_base, filename))
        data = json.load(f)
        f.close()
        return data

    def test__verify_cert(self):
        verify_cert = self.client.verify_certificate
        certificate_path = self.client.certificate_path
        self.client.verify_certificate = True
        self.client.certificate_path = "fake_certificate_path"
        self.assertEqual(self.client._verify_cert,
                         self.client.certificate_path)
        self.client.verify_certificate = verify_cert
        self.client.certificate_path = certificate_path

    @requests_mock.mock()
    def test__send_request(self, m):
        url = "{0}/fake_res".format(self._mock_url)
        m.get(url, status_code=200)
        self.client._send_get_request("/fake_res", None, None, False)

    @requests_mock.mock()
    def test_get_nas_server_id(self, m):
        self.assertEqual(0, len(m.request_history))
        self._add_get_nas_server_id_response(
            m, self.NAS_SERVER_NAME,
            self._getJsonFile("get_nas_server_id_response.json")
        )
        id = self.client.get_nas_server_id(self.NAS_SERVER_NAME)
        self.assertEqual(id, self.NAS_SERVER_ID)

    def _add_get_nas_server_id_response(self, m, nas_server, json_str):
        url = "{0}/nas_server?name=eq.{1}".format(
            self._mock_url, nas_server
        )
        m.get(url, status_code=200, json=json_str)

    @requests_mock.mock()
    def test_get_nas_server_id_failure(self, m):
        self.assertEqual(0, len(m.request_history))
        self._add_get_nas_server_id_response_failure(
            m, self.NAS_SERVER_NAME
        )
        id = self.client.get_nas_server_id(self.NAS_SERVER_NAME)
        self.assertIsNone(id)

    def _add_get_nas_server_id_response_failure(self, m, nas_server):
        url = "{0}/nas_server?name=eq.{1}".format(
            self._mock_url, nas_server
        )
        m.get(url, status_code=400)

    @requests_mock.mock()
    def test_get_nas_server_interfaces(self, m):
        self.assertEqual(0, len(m.request_history))
        self._add_get_nas_server_interfaces_response(
            m, self.NAS_SERVER_ID,
            self._getJsonFile("get_nas_server_interfaces_response.json")
        )
        interfaces = self.client.get_nas_server_interfaces(self.NAS_SERVER_ID)
        self.assertEqual(interfaces[0]['ip'], self.NAS_SERVER_IP)
        self.assertEqual(interfaces[0]['preferred'], True)

    def _add_get_nas_server_interfaces_response(self, m, nas_server_id,
                                                json_str):
        url = "{0}/nas_server/{1}?select=" \
            "current_preferred_IPv4_interface_id," \
            "current_preferred_IPv6_interface_id," \
            "file_interfaces(id,ip_address)".format(
                self._mock_url, nas_server_id
            )
        m.get(url, status_code=200, json=json_str)

    @requests_mock.mock()
    def test_get_nas_server_interfaces_failure(self, m):
        self.assertEqual(0, len(m.request_history))
        self._add_get_nas_server_interfaces_response_failure(
            m, self.NAS_SERVER_ID
        )
        interfaces = self.client.get_nas_server_interfaces(self.NAS_SERVER_ID)
        self.assertIsNone(interfaces)

    def _add_get_nas_server_interfaces_response_failure(self, m,
                                                        nas_server_id):
        url = "{0}/nas_server/{1}?select=" \
            "current_preferred_IPv4_interface_id," \
            "current_preferred_IPv6_interface_id," \
            "file_interfaces(id,ip_address)".format(
                self._mock_url, nas_server_id
            )
        m.get(url, status_code=400)

    @requests_mock.mock()
    def test_create_filesystem(self, m):
        self.assertEqual(0, len(m.request_history))
        self._add_create_filesystem_response(
            m, self._getJsonFile("create_filesystem_response.json")
        )
        id = self.client.create_filesystem(
            self.NAS_SERVER_ID,
            self.NFS_EXPORT_NAME,
            self.NFS_EXPORT_SIZE
        )
        self.assertEqual(id, self.FILESYSTEM_ID)

    def _add_create_filesystem_response(self, m, json_str):
        url = "{0}/file_system".format(self._mock_url)
        m.post(url, status_code=201, json=json_str)

    @requests_mock.mock()
    def test_create_filesystem_failure(self, m):
        self.assertEqual(0, len(m.request_history))
        self._add_create_filesystem_response_failure(m)
        id = self.client.create_filesystem(
            self.NAS_SERVER_ID,
            self.NFS_EXPORT_NAME,
            self.NFS_EXPORT_SIZE
        )
        self.assertIsNone(id)

    def _add_create_filesystem_response_failure(self, m):
        url = "{0}/file_system".format(self._mock_url)
        m.post(url, status_code=400)

    @requests_mock.mock()
    def test_create_nfs_export(self, m):
        self.assertEqual(0, len(m.request_history))
        self._add_create_nfs_export_response(
            m, self._getJsonFile("create_nfs_export_response.json")
        )
        id = self.client.create_nfs_export(self.FILESYSTEM_ID,
                                           self.NFS_EXPORT_NAME)
        self.assertEqual(id, self.NFS_EXPORT_ID)

    def _add_create_nfs_export_response(self, m, json_str):
        url = "{0}/nfs_export".format(self._mock_url)
        m.post(url, status_code=201, json=json_str)

    @requests_mock.mock()
    def test_create_nfs_export_failure(self, m):
        self.assertEqual(0, len(m.request_history))
        self._add_create_nfs_export_response_failure(m)
        id = self.client.create_nfs_export(self.FILESYSTEM_ID,
                                           self.NFS_EXPORT_NAME)
        self.assertIsNone(id)

    def _add_create_nfs_export_response_failure(self, m):
        url = "{0}/nfs_export".format(self._mock_url)
        m.post(url, status_code=400)

    @requests_mock.mock()
    def test_delete_filesystem(self, m):
        self.assertEqual(0, len(m.request_history))
        self._add_delete_filesystem_response(m, self.FILESYSTEM_ID)
        result = self.client.delete_filesystem(self.FILESYSTEM_ID)
        self.assertEqual(result, True)

    def _add_delete_filesystem_response(self, m, filesystem_id):
        url = "{0}/file_system/{1}".format(
            self._mock_url, filesystem_id
        )
        m.delete(url, status_code=204)

    @requests_mock.mock()
    def test_get_nfs_export_name(self, m):
        self.assertEqual(0, len(m.request_history))
        self._add_get_nfs_export_name_response(
            m,
            self.NFS_EXPORT_ID,
            self._getJsonFile("get_nfs_export_name_response.json"),
        )
        name = self.client.get_nfs_export_name(self.NFS_EXPORT_ID)
        self.assertEqual(name, self.NFS_EXPORT_NAME)

    def _add_get_nfs_export_name_response(self, m, export_id, json_str):
        url = "{0}/nfs_export/{1}?select=name".format(
            self._mock_url, export_id
        )
        m.get(url, status_code=200, json=json_str)

    @requests_mock.mock()
    def test_get_nfs_export_name_failure(self, m):
        self.assertEqual(0, len(m.request_history))
        self._add_get_nfs_export_name_response_failure(m, self.NFS_EXPORT_ID)
        name = self.client.get_nfs_export_name(self.NFS_EXPORT_ID)
        self.assertIsNone(name)

    def _add_get_nfs_export_name_response_failure(self,
                                                  m, export_id):
        url = "{0}/nfs_export/{1}?select=name".format(
            self._mock_url, export_id
        )
        m.get(url, status_code=400)

    @requests_mock.mock()
    def test_get_nfs_export_id(self, m):
        self.assertEqual(0, len(m.request_history))
        self._add_get_nfs_export_id_response(
            m, self.NFS_EXPORT_NAME,
            self._getJsonFile("get_nfs_export_id_response.json")
        )
        id = self.client.get_nfs_export_id(self.NFS_EXPORT_NAME)
        self.assertEqual(id, self.NFS_EXPORT_ID)

    def _add_get_nfs_export_id_response(self, m, name, json_str):
        url = "{0}/nfs_export?select=id&name=eq.{1}".format(
            self._mock_url, name
        )
        m.get(url, status_code=200, json=json_str)

    @requests_mock.mock()
    def test_get_nfs_export_id_failure(self, m):
        self.assertEqual(0, len(m.request_history))
        self._add_get_nfs_export_id_response_failure(m, self.NFS_EXPORT_NAME)
        id = self.client.get_nfs_export_id(self.NFS_EXPORT_NAME)
        self.assertIsNone(id)

    def _add_get_nfs_export_id_response_failure(self, m, name):
        url = "{0}/nfs_export?select=id&name=eq.{1}".format(
            self._mock_url, name
        )
        m.get(url, status_code=400)

    @requests_mock.mock()
    def test_get_filesystem_id(self, m):
        self.assertEqual(0, len(m.request_history))
        self._add_get_filesystem_id_response(
            m, self.NFS_EXPORT_NAME,
            self._getJsonFile("get_fileystem_id_response.json")
        )
        id = self.client.get_filesystem_id(self.NFS_EXPORT_NAME)
        self.assertEqual(id, self.FILESYSTEM_ID)

    def _add_get_filesystem_id_response(self, m, name, json_str):
        url = "{0}/file_system?name=eq.{1}".format(
            self._mock_url, name
        )
        m.get(url, status_code=200, json=json_str)

    @requests_mock.mock()
    def test_get_filesystem_id_failure(self, m):
        self.assertEqual(0, len(m.request_history))
        self._add_get_filesystem_id_response_failure(m, self.NFS_EXPORT_NAME)
        id = self.client.get_filesystem_id(self.NFS_EXPORT_NAME)
        self.assertIsNone(id)

    def _add_get_filesystem_id_response_failure(self, m, name):
        url = "{0}/file_system?name=eq.{1}".format(
            self._mock_url, name
        )
        m.get(url, status_code=400)

    @requests_mock.mock()
    def test_set_export_access(self, m):

        self.assertEqual(0, len(m.request_history))
        self._add_set_export_access_response(m, self.NFS_EXPORT_ID)
        result = self.client.set_export_access(self.NFS_EXPORT_ID,
                                               self.RW_HOSTS,
                                               self.RO_HOSTS)
        self.assertEqual(result, True)

    def _add_set_export_access_response(self, m, export_id):
        url = "{0}/nfs_export/{1}".format(self._mock_url, export_id)
        m.patch(url, status_code=204)

    @requests_mock.mock()
    def test_resize_filesystem(self, m):
        self.assertEqual(0, len(m.request_history))
        self._add_resize_filesystem_response(m, self.FILESYSTEM_ID)
        result, detail = self.client.resize_filesystem(
            self.FILESYSTEM_ID, self.NFS_EXPORT_NEW_SIZE)
        self.assertTrue(result)
        self.assertIsNone(detail)

    def _add_resize_filesystem_response(self, m, filesystem_id):
        url = "{0}/file_system/{1}".format(
            self._mock_url, filesystem_id
        )
        m.patch(url, status_code=204)

    @requests_mock.mock()
    def test_resize_filesystem_shrink_failure(self, m):
        self.assertEqual(0, len(m.request_history))
        self._add_resize_filesystem_shrink_failure_response(
            m, self.FILESYSTEM_ID,
            self._getJsonFile(
                "resize_filesystem_shrink_failure_response.json"))
        result, detail = self.client.resize_filesystem(
            self.FILESYSTEM_ID, self.NFS_EXPORT_NEW_SIZE)
        self.assertFalse(result)
        self.assertIsNotNone(detail)

    def _add_resize_filesystem_shrink_failure_response(
            self, m, filesystem_id, json_str):
        url = "{0}/file_system/{1}".format(
            self._mock_url, filesystem_id
        )
        m.patch(url, status_code=422, json=json_str)

    @requests_mock.mock()
    def test_get_fsid_from_export_name(self, m):
        self.assertEqual(0, len(m.request_history))
        self._add_get_fsid_from_export_name_response(
            m, self.NFS_EXPORT_NAME,
            self._getJsonFile("get_fsid_from_export_name_response.json")
        )
        id = self.client.get_fsid_from_export_name(self.NFS_EXPORT_NAME)
        self.assertEqual(id, self.FILESYSTEM_ID)

    def _add_get_fsid_from_export_name_response(self, m, name, json_str):
        url = "{0}/nfs_export?select=file_system_id&name=eq.{1}".format(
            self._mock_url, name
        )
        m.get(url, status_code=200, json=json_str)

    @requests_mock.mock()
    def test_get_fsid_from_export_name_failure(self, m):
        self.assertEqual(0, len(m.request_history))
        self._add_get_fsid_from_export_name_response_failure(
            m, self.NFS_EXPORT_NAME
        )
        id = self.client.get_fsid_from_export_name(self.NFS_EXPORT_NAME)
        self.assertIsNone(id)

    def _add_get_fsid_from_export_name_response_failure(self, m, name):
        url = "{0}/nfs_export?select=file_system_id&name=eq.{1}".format(
            self._mock_url, name
        )
        m.get(url, status_code=400)

    @requests_mock.mock()
    def test_create_snapshot(self, m):
        self.assertEqual(0, len(m.request_history))
        self._add_create_snapshot_response(
            m, self.FILESYSTEM_ID,
            self._getJsonFile("create_snapshot_response.json")
        )
        id = self.client.create_snapshot(self.FILESYSTEM_ID,
                                         self.SNAPSHOT_NAME)
        self.assertEqual(id, self.SNAPSHOT_ID)

    def _add_create_snapshot_response(self, m, filesystem_id, json_str):
        url = "{0}/file_system/{1}/snapshot".format(self._mock_url,
                                                    filesystem_id)
        m.post(url, status_code=201, json=json_str)

    @requests_mock.mock()
    def test_create_snapshot_failure(self, m):
        self.assertEqual(0, len(m.request_history))
        self._add_create_snapshot_response_failure(m, self.FILESYSTEM_ID)
        id = self.client.create_snapshot(self.FILESYSTEM_ID,
                                         self.SNAPSHOT_NAME)
        self.assertIsNone(id)

    def _add_create_snapshot_response_failure(self, m, filesystem_id):
        url = "{0}/file_system/{1}/snapshot".format(self._mock_url,
                                                    filesystem_id)
        m.post(url, status_code=400)

    @requests_mock.mock()
    def test_restore_snapshot(self, m):
        self.assertEqual(0, len(m.request_history))
        self._add_restore_snapshot_response(
            m, self.SNAPSHOT_ID
        )
        result = self.client.restore_snapshot(self.SNAPSHOT_ID)
        self.assertEqual(result, True)

    def _add_restore_snapshot_response(self, m, snapshot_id):
        url = "{0}/file_system/{1}/restore".format(self._mock_url,
                                                   snapshot_id)
        m.post(url, status_code=204)

    @requests_mock.mock()
    def test_restore_snapshot_failure(self, m):
        self.assertEqual(0, len(m.request_history))
        self._add_restore_snapshot_response_failure(
            m, self.SNAPSHOT_ID
        )
        result = self.client.restore_snapshot(self.SNAPSHOT_ID)
        self.assertEqual(result, False)

    def _add_restore_snapshot_response_failure(self, m, snapshot_id):
        url = "{0}/file_system/{1}/restore".format(self._mock_url,
                                                   snapshot_id)
        m.post(url, status_code=400)

    @requests_mock.mock()
    def test_clone_snapshot(self, m):
        self.assertEqual(0, len(m.request_history))
        self._add_clone_snapshot_response(
            m, self.SNAPSHOT_ID,
            self._getJsonFile("clone_snapshot_response.json")
        )
        id = self.client.clone_snapshot(self.SNAPSHOT_ID,
                                        self.CLONE_NAME)
        self.assertEqual(id, self.CLONE_ID)

    def _add_clone_snapshot_response(self, m, snapshot_id, json_str):
        url = "{0}/file_system/{1}/clone".format(self._mock_url,
                                                 snapshot_id)
        m.post(url, status_code=201, json=json_str)

    @requests_mock.mock()
    def test_clone_snapshot_failure(self, m):
        self.assertEqual(0, len(m.request_history))
        self._add_clone_snapshot_response_failure(
            m, self.SNAPSHOT_ID
        )
        id = self.client.clone_snapshot(self.SNAPSHOT_ID,
                                        self.CLONE_NAME)
        self.assertIsNone(id)

    def _add_clone_snapshot_response_failure(self, m, snapshot_id):
        url = "{0}/file_system/{1}/clone".format(self._mock_url,
                                                 snapshot_id)
        m.post(url, status_code=400)

    @requests_mock.mock()
    def test_get_cluster_id(self, m):
        self.assertEqual(0, len(m.request_history))
        self._add_get_cluster_id_response(
            m,
            self._getJsonFile("get_cluster_id_response.json")
        )
        id = self.client.get_cluster_id()
        self.assertEqual(id, self.CLUSTER_ID)

    def _add_get_cluster_id_response(self, m, json_str):
        url = "{0}/cluster".format(self._mock_url)
        m.get(url, status_code=200, json=json_str)

    @requests_mock.mock()
    def test_get_cluster_id_failure(self, m):
        self.assertEqual(0, len(m.request_history))
        self._add_get_cluster_id_response_failure(m)
        id = self.client.get_cluster_id()
        self.assertIsNone(id)

    def _add_get_cluster_id_response_failure(self, m):
        url = "{0}/cluster".format(self._mock_url)
        m.get(url, status_code=400)

    @requests_mock.mock()
    def test_retreive_cluster_capacity_metrics(self, m):
        self.assertEqual(0, len(m.request_history))
        self._add_retreive_cluster_capacity_metrics_response(
            m,
            self._getJsonFile(
                "retreive_cluster_capacity_metrics_response.json")
        )
        total, used = self.client.retreive_cluster_capacity_metrics(
            self.CLUSTER_ID)
        self.assertEqual(total, 47345047046144)
        self.assertEqual(used, 366003363027)

    def _add_retreive_cluster_capacity_metrics_response(self, m, json_str):
        url = "{0}/metrics/generate?order=timestamp".format(self._mock_url)
        m.post(url, status_code=200, json=json_str)

    @requests_mock.mock()
    def test_retreive_cluster_capacity_metrics_failure(self, m):
        self.assertEqual(0, len(m.request_history))
        self._add_retreive_cluster_capacity_metrics_response_failure(m)
        total, used = self.client.retreive_cluster_capacity_metrics(
            self.CLUSTER_ID)
        self.assertIsNone(total)
        self.assertIsNone(used)

    def _add_retreive_cluster_capacity_metrics_response_failure(self, m):
        url = "{0}/metrics/generate?order=timestamp".format(self._mock_url)
        m.post(url, status_code=400)

    @requests_mock.mock()
    def test_create_smb_share(self, m):
        self.assertEqual(0, len(m.request_history))
        self._add_create_smb_share_response(
            m, self._getJsonFile("create_smb_share_response.json")
        )
        id = self.client.create_smb_share(self.FILESYSTEM_ID,
                                          self.SMB_SHARE_NAME)
        self.assertEqual(id, self.SMB_SHARE_ID)

    def _add_create_smb_share_response(self, m, json_str):
        url = "{0}/smb_share".format(self._mock_url)
        m.post(url, status_code=201, json=json_str)

    @requests_mock.mock()
    def test_create_smb_share_failure(self, m):
        self.assertEqual(0, len(m.request_history))
        self._add_create_smb_share_response_failure(m)
        id = self.client.create_smb_share(self.FILESYSTEM_ID,
                                          self.SMB_SHARE_NAME)
        self.assertIsNone(id)

    def _add_create_smb_share_response_failure(self, m):
        url = "{0}/smb_share".format(self._mock_url)
        m.post(url, status_code=400)

    @requests_mock.mock()
    def test_get_fsid_from_share_name(self, m):
        self.assertEqual(0, len(m.request_history))
        self._add_get_fsid_from_share_name_response(
            m, self.NFS_EXPORT_NAME,
            self._getJsonFile("get_fsid_from_share_name_response.json")
        )
        id = self.client.get_fsid_from_share_name(self.NFS_EXPORT_NAME)
        self.assertEqual(id, self.FILESYSTEM_ID)

    def _add_get_fsid_from_share_name_response(self, m, name, json_str):
        url = "{0}/smb_share?select=file_system_id&name=eq.{1}".format(
            self._mock_url, name
        )
        m.get(url, status_code=200, json=json_str)

    @requests_mock.mock()
    def test_get_fsid_from_share_name_failure(self, m):
        self.assertEqual(0, len(m.request_history))
        self._add_get_fsid_from_share_name_response_failure(
            m, self.SMB_SHARE_NAME
        )
        id = self.client.get_fsid_from_share_name(self.SMB_SHARE_NAME)
        self.assertIsNone(id)

    def _add_get_fsid_from_share_name_response_failure(self, m, name):
        url = "{0}/smb_share?select=file_system_id&name=eq.{1}".format(
            self._mock_url, name
        )
        m.get(url, status_code=400)

    @requests_mock.mock()
    def test_get_smb_share_id(self, m):
        self.assertEqual(0, len(m.request_history))
        self._add_get_smb_share_id_response(
            m, self.SMB_SHARE_NAME,
            self._getJsonFile("get_smb_share_id_response.json")
        )
        id = self.client.get_smb_share_id(self.SMB_SHARE_NAME)
        self.assertEqual(id, self.SMB_SHARE_ID)

    def _add_get_smb_share_id_response(self, m, name, json_str):
        url = "{0}/smb_share?select=id&name=eq.{1}".format(
            self._mock_url, name
        )
        m.get(url, status_code=200, json=json_str)

    @requests_mock.mock()
    def test_get_smb_share_id_failure(self, m):
        self.assertEqual(0, len(m.request_history))
        self._add_get_smb_share_id_response_failure(
            m, self.SMB_SHARE_NAME
        )
        id = self.client.get_smb_share_id(self.SMB_SHARE_NAME)
        self.assertIsNone(id)

    def _add_get_smb_share_id_response_failure(self, m, name):
        url = "{0}/smb_share?select=id&name=eq.{1}".format(
            self._mock_url, name
        )
        m.get(url, status_code=400)

    @requests_mock.mock()
    def test_get_nas_server_smb_netbios(self, m):
        self.assertEqual(0, len(m.request_history))
        self._add_get_nas_server_smb_netbios_response(
            m, self.SMB_SHARE_NAME,
            self._getJsonFile("get_nas_server_smb_netbios_response.json")
        )
        id = self.client.get_nas_server_smb_netbios(self.SMB_SHARE_NAME)
        self.assertEqual(id, "OPENSTACK")

    def _add_get_nas_server_smb_netbios_response(self, m, name, json_str):
        url = "{0}/nas_server?select=smb_servers" \
            "(is_standalone,netbios_name)&name=eq.{1}".format(
                self._mock_url, name
            )
        m.get(url, status_code=200, json=json_str)

    @requests_mock.mock()
    def test_get_nas_server_smb_netbios_failure(self, m):
        self.assertEqual(0, len(m.request_history))
        self._add_get_nas_server_smb_netbios_response_failure(
            m, self.SMB_SHARE_NAME
        )
        id = self.client.get_nas_server_smb_netbios(self.SMB_SHARE_NAME)
        self.assertIsNone(id)

    def _add_get_nas_server_smb_netbios_response_failure(self, m, name):
        url = "{0}/nas_server?select=smb_servers" \
            "(is_standalone,netbios_name)&name=eq.{1}".format(
                self._mock_url, name
            )
        m.get(url, status_code=400)

    @requests_mock.mock()
    def test_set_acl(self, m):

        self.assertEqual(0, len(m.request_history))
        self._add_set_acl_response(m, self.SMB_SHARE_ID)
        result = self.client.set_acl(self.SMB_SHARE_ID,
                                     self.RW_USERS,
                                     self.RO_USERS)
        self.assertEqual(result, True)

    def _add_set_acl_response(self, m, share_id):
        url = "{0}/smb_share/{1}/set_acl".format(self._mock_url, share_id)
        m.post(url, status_code=204)



# Copyright 2023 NetApp, Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
NetApp API for REST Data ONTAP.

Contains classes required to issue REST API calls to Data ONTAP.
"""

import re

from oslo_log import log
from oslo_serialization import jsonutils
import requests
from requests.adapters import HTTPAdapter
from requests import auth
from urllib3.util import retry

from manila.share.drivers.netapp.dataontap.client import api
from manila.share.drivers.netapp import utils


LOG = log.getLogger(__name__)

EREST_DUPLICATE_ENTRY = '1'
EREST_ENTRY_NOT_FOUND = '4'
EREST_NOT_AUTHORIZED = '6'
EREST_SNAPMIRROR_INITIALIZING = '917536'
EREST_VSERVER_NOT_FOUND = '13434920'
EREST_ANOTHER_VOLUME_OPERATION = '13107406'
EREST_LICENSE_NOT_INSTALLED = '1115127'
EREST_SNAPSHOT_NOT_SPECIFIED = '1638515'
EREST_FPOLICY_MODIF_POLICY_DISABLED = '9765029'
EREST_POLICY_ALREADY_DISABLED = '9764907'
EREST_ERELATION_EXISTS = '6619637'
EREST_BREAK_SNAPMIRROR_FAILED = '13303808'
EREST_UPDATE_SNAPMIRROR_FAILED = '13303844'
EREST_SNAPMIRROR_NOT_INITIALIZED = '13303812'
EREST_DUPLICATE_ROUTE = '1966345'
EREST_FAIL_ADD_PORT_BROADCAST = '1967149'
EREST_KERBEROS_IS_ENABLED_DISABLED = '3276861'
EREST_INTERFACE_BOUND = '1376858'
EREST_PORT_IN_USE = '1966189'
EREST_NFS_V4_0_ENABLED_MIGRATION_FAILURE = '13172940'
EREST_VSERVER_MIGRATION_TO_NON_AFF_CLUSTER = '13172984'
EREST_UNMOUNT_FAILED_LOCK = '917536'
EREST_CANNOT_MODITY_OFFLINE_VOLUME = '917533'


class NaRetryableError(api.NaApiError):
    def __str__(self, *args, **kwargs):
        return 'NetApp API failed. Try again. Reason - %s:%s' % (
            self.code, self.message)


class RestNaServer(object):

    TRANSPORT_TYPE_HTTP = 'http'
    TRANSPORT_TYPE_HTTPS = 'https'
    HTTP_PORT = '80'
    HTTPS_PORT = '443'
    TUNNELING_HEADER_KEY = "X-Dot-SVM-Name"

    def __init__(self, host, transport_type=TRANSPORT_TYPE_HTTP,
                 ssl_cert_path=None, username=None, password=None, port=None,
                 trace=False, api_trace_pattern=utils.API_TRACE_PATTERN):
        self._host = host
        self.set_transport_type(transport_type)
        self.set_port(port=port)
        self._username = username
        self._password = password
        self._trace = trace
        self._api_trace_pattern = api_trace_pattern
        self._timeout = None

        if ssl_cert_path is not None:
            self._ssl_verify = ssl_cert_path
        else:
            # Note(felipe_rodrigues): it will verify with the mozila CA roots,
            # given by certifi package.
            self._ssl_verify = True

        LOG.debug('Using REST with NetApp controller: %s', self._host)

    def set_transport_type(self, transport_type):
        """Set the transport type protocol for API.

        Supports http and https transport types.
        """
        if transport_type is None or transport_type.lower() not in (
                RestNaServer.TRANSPORT_TYPE_HTTP,
                RestNaServer.TRANSPORT_TYPE_HTTPS):
            raise ValueError('Unsupported transport type')
        self._protocol = transport_type.lower()

    def get_transport_type(self):
        """Get the transport type protocol."""
        return self._protocol

    def set_api_version(self, major, minor):
        """Set the API version."""
        try:
            self._api_major_version = int(major)
            self._api_minor_version = int(minor)
            self._api_version = str(major) + "." + str(minor)
        except ValueError:
            raise ValueError('Major and minor versions must be integers')

    def get_api_version(self):
        """Gets the API version tuple."""
        if hasattr(self, '_api_version'):
            return (self._api_major_version, self._api_minor_version)
        return None

    def set_ontap_version(self, ontap_version):
        """Set the ONTAP version."""
        self._ontap_version = ontap_version

    def get_ontap_version(self):
        """Gets the ONTAP version."""
        if hasattr(self, '_ontap_version'):
            return self._ontap_version
        return None

    def set_port(self, port=None):
        """Set the ONTAP port, if not informed, set with default one."""
        if port is None and self._protocol == RestNaServer.TRANSPORT_TYPE_HTTP:
            self._port = RestNaServer.HTTP_PORT
        elif port is None:
            self._port = RestNaServer.HTTPS_PORT
        else:
            try:
                int(port)
            except ValueError:
                raise ValueError('Port must be integer')
            self._port = str(port)

    def get_port(self):
        """Get the server communication port."""
        return self._port

    def set_timeout(self, seconds):
        """Sets the timeout in seconds."""
        try:
            self._timeout = int(seconds)
        except ValueError:
            raise ValueError('timeout in seconds must be integer')

    def get_timeout(self):
        """Gets the timeout in seconds if set."""
        return self._timeout

    def set_vserver(self, vserver):
        """Set the vserver to use if tunneling gets enabled."""
        self._vserver = vserver

    def get_vserver(self):
        """Get the vserver to use in tunneling."""
        return self._vserver

    def __str__(self):
        """Gets a representation of the client."""
        return "server: %s" % (self._host)

    def _get_request_method(self, method, session):
        """Returns the request method to be used in the REST call."""

        request_methods = {
            'post': session.post,
            'get': session.get,
            'put': session.put,
            'delete': session.delete,
            'patch': session.patch,
        }
        return request_methods[method]

    def _add_query_params_to_url(self, url, query):
        """Populates the URL with specified filters."""
        filters = '&'.join([f"{k}={v}" for k, v in query.items()])
        url += "?" + filters
        return url

    def _get_base_url(self):
        """Get the base URL for REST requests."""
        host = self._host
        if ':' in host:
            host = '[%s]' % host
        return f'{self._protocol}://{host}:{self._port}/api'

    def _build_session(self, headers):
        """Builds a session in the client."""
        self._session = requests.Session()

        # NOTE(felipe_rodrigues): request resilient of temporary network
        # failures (like name resolution failure), retrying until 5 times.
        max_retries = retry.Retry(total=5, connect=5, read=2, backoff_factor=1)
        adapter = HTTPAdapter(max_retries=max_retries)
        self._session.mount('%s://' % self._protocol, adapter)

        self._session.auth = self._create_basic_auth_handler()
        self._session.verify = self._ssl_verify
        self._session.headers = headers

    def _build_headers(self, enable_tunneling):
        """Build and return headers for a REST request."""
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json"
        }
        # enable tunneling only if vserver is set by upper layer
        if enable_tunneling and self.get_vserver:
            headers[RestNaServer.TUNNELING_HEADER_KEY] = self.get_vserver()

        return headers

    def _create_basic_auth_handler(self):
        """Creates and returns a basic HTTP auth handler."""
        return auth.HTTPBasicAuth(self._username, self._password)

    def send_http_request(self, method, url, body, headers):
        """Invoke the API on the server."""
        data = jsonutils.dumps(body) if body else {}

        self._build_session(headers)
        request_method = self._get_request_method(method, self._session)

        api_name_matches_regex = (re.match(self._api_trace_pattern, url)
                                  is not None)
        if self._trace and api_name_matches_regex:
            svm = headers.get(RestNaServer.TUNNELING_HEADER_KEY)
            message = ("Request: %(method)s Header=%(header)s %(url)s "
                       "Body=%(body)s")
            msg_args = {
                "method": method.upper(),
                "url": url,
                "body": body,
                "header": ({RestNaServer.TUNNELING_HEADER_KEY: svm}
                           if svm else {}),
            }
            LOG.debug(message, msg_args)

        try:
            if self._timeout is not None:
                response = request_method(
                    url, data=data, timeout=self._timeout)
            else:
                response = request_method(url, data=data)
        except requests.HTTPError as e:
            raise api.NaApiError(e.errno, e.strerror)
        except Exception as e:
            raise api.NaApiError(message=e)

        code = response.status_code
        res = jsonutils.loads(response.content) if response.content else {}

        if self._trace and api_name_matches_regex:
            message = "Response: %(code)s Body=%(body)s"
            msg_args = {
                "code": code,
                "body": res
            }
            LOG.debug(message, msg_args)

        return code, res

    def invoke_successfully(self, action_url, method, body=None, query=None,
                            enable_tunneling=False):
        """Invokes REST API and checks execution status as success."""
        headers = self._build_headers(enable_tunneling)
        if query:
            action_url = self._add_query_params_to_url(action_url, query)
        url = self._get_base_url() + action_url
        code, response = self.send_http_request(method, url, body, headers)

        if not response.get('error'):
            return code, response

        result_error = response.get('error')
        code = result_error.get('code') or 'ESTATUSFAILED'
        msg = (result_error.get('message')
               or 'Execution failed due to unknown reason')
        raise api.NaApiError(code, msg)



# Copyright (c) 2022 MacroSAN Technologies Co., Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import requests

from oslo_log import log

from manila import exception
from manila.i18n import _
from manila.share.drivers.macrosan import macrosan_constants as constants
from manila import utils

LOG = log.getLogger(__name__)


class RestHelper(object):
    def __init__(self, configuration):
        self.configuration = configuration
        self._protocol = self.configuration.macrosan_nas_http_protocol
        self._ip = self.configuration.macrosan_nas_ip
        self._port = self.configuration.macrosan_nas_port
        self._prefix = self.configuration.macrosan_nas_prefix
        self._token = None
        self._username = self.configuration.macrosan_nas_username
        self._password = self.configuration.macrosan_nas_password
        self.request_timeout = self.configuration.macrosan_timeout
        self.ssl_verify = self.configuration.macrosan_ssl_cert_verify
        if not self.ssl_verify:
            # Suppress the Insecure request warnings
            requests.packages.urllib3.disable_warnings(
                requests.packages.urllib3.exceptions.InsecureRequestWarning)

    @utils.synchronized('macrosan_manila')
    def call(self, url, data, method):
        """Send requests.

        If token is expired,re-login.
        """
        header = {'Authorization': self._token}
        if self._token is None:
            self.login()

        result = self.do_request(url, data, method, header)
        if result['code'] == constants.TOKEN_EXPIRED:
            LOG.error("Token is expired, re-login.")
            self.login()
            # token refresh, Re-assign
            header['Authorization'] = self._token
            result = self.do_request(url, data, method, header)
        elif (result['code'] == constants.TOKEN_FORMAT_ERROR or
              result['code'] == constants.TOKEN_VERIFY_FAILED or
              result['code'] == constants.TOKEN_REQUIRED):
            msg = _('Token authentication error.')
            LOG.error(msg)
            raise exception.MacrosanBackendExeption(msg)
        return result

    def do_request(self, url, data, method, header=None):
        final_url = (f'{self._protocol}://{self._ip}:{self._port}/'
                     f'{self._prefix}/{url}')
        LOG.debug(f'Request URL: {final_url}, Method: {method}, Data: {data}')

        if method == 'POST':
            res = requests.post(final_url, data=data, headers=header,
                                timeout=self.request_timeout,
                                verify=self.ssl_verify)
        elif method == 'GET':
            res = requests.get(final_url, data=data, headers=header,
                               timeout=self.request_timeout,
                               verify=self.ssl_verify)
        elif method == 'PUT':
            res = requests.put(final_url, data=data, headers=header,
                               timeout=self.request_timeout,
                               verify=self.ssl_verify)
        elif method == 'DELETE':
            res = requests.delete(final_url, data=data, headers=header,
                                  timeout=self.request_timeout,
                                  verify=self.ssl_verify)
        else:
            msg = (_("Request method %s invalid.") % method)
            raise exception.ShareBackendException(msg=msg)

        code = res.status_code
        if code != 200:
            msg = (_('Code: %(code)s, URL: %(url)s, Message: %(msg)s')
                   % {'code': res.status_code,
                      'url': final_url,
                      'msg': res.text})
            LOG.error(msg)
            raise exception.NetworkException(msg)
        response = res.json()
        LOG.debug('CODE: %(code)s, RESPONSE: %(response)s',
                  {'code': code, 'response': response})
        return response

    def login(self):
        """Login array and return token."""
        url = 'rest/token'

        data = {'userName': self._username,
                'userPasswd': self._password}
        result = self.do_request(url, data, 'POST')
        if result['code'] != 0:
            msg = f"Login failed. code: {result['code']}"
            msg = _(msg)
            LOG.error(msg)
            raise exception.ShareBackendException(msg=msg)
        LOG.debug(f'Login successful. URL {self._ip}\n')
        self._token = result['data']

    def _assert_result_code(self, result, msg):
        if (result['code'] != constants.CODE_SUCCESS
                and result['code'] != constants.CODE_NOT_FOUND):
            error_msg = (_('%(err)s\nresult: %(res)s.') % {'err': msg,
                                                           'res': result})
            LOG.error(error_msg)
            raise exception.ShareBackendException(msg=error_msg)

    def _assert_result_data(self, result, msg):
        if "data" not in result:
            error_msg = (_('Error:"data" not in result. %s') % msg)
            LOG.error(error_msg)
            raise exception.ShareBackendException(msg=error_msg)

    def _create_nfs_share(self, share_path):
        url = 'rest/nfsShare'
        # IPv4 Address Blocks Reserved for Documentation
        params = {
            'path': share_path,
            'authority': 'ro',
            'accessClient': '192.0.2.0',
        }
        result = self.call(url, params, 'POST')

        msg = 'Failed to create a nfs share.'
        self._assert_result_code(result, msg)

    def _get_nfs_share(self, share_path):
        # GET method: param need be after url
        url = f'rest/nfsShare?path={share_path}'
        result = self.call(url, None, 'GET')

        msg = 'Failed to get nfs share.'
        self._assert_result_code(result, msg)
        return result['data']

    def _delete_nfs_share(self, share_path):
        url = f'rest/nfsShare?path={share_path}'
        result = self.call(url, None, 'DELETE')

        msg = 'Failed to delete nfs share.'
        self._assert_result_code(result, msg)

    def _create_cifs_share(self, share_name, share_path,
                           rw_list, rw_list_type):
        url = 'rest/cifsShare'

        params = {
            'path': share_path,
            'cifsName': share_name,
            'cifsDescription': '',
            'RoList': [],
            'RoListType': [],
            'RwList': rw_list,
            'RwListType': rw_list_type,
            'allowList': [],
            'denyList': [],
        }
        result = self.call(url, params, 'POST')

        msg = 'Failed to create a CIFS share.'
        self._assert_result_code(result, msg)

    def _get_cifs_share(self, share_path):
        url = f'rest/cifsShare?path={share_path}'
        result = self.call(url, None, 'GET')

        msg = 'Failed to get the cifs share.'
        self._assert_result_code(result, msg)
        return result['data']

    def _delete_cifs_share(self, share_name, share_path):
        url = f'rest/cifsShare?path={share_path}&cifsName={share_name}'
        result = self.call(url, None, 'DELETE')
        msg = 'Failed to delete the cifs share.'

        self._assert_result_code(result, msg)

    def _update_share_size(self, fs_name, new_size):
        url = f'rest/filesystem/{fs_name}'

        params = {
            'capacity': new_size,
        }

        result = self.call(url, params, 'PUT')
        msg = 'Failed to update the filesystem size.'

        self._assert_result_code(result, msg)

    def _create_filesystem(self, fs_name, pool_name, filesystem_quota):
        url = 'rest/filesystem'
        fsinfo = {
            'fsName': fs_name,
            'poolName': pool_name,
            'createType': '0',
            'fileSystemQuota': filesystem_quota,
            'fileSystemReserve': filesystem_quota,
            'wormStatus': 0,
            'defaultTimeStatus': 0,
            'defaultTimeNum': 0,
            'defaultTimeUnit': 'year',
            'isAutoLock': 0,
            'isAutoDelete': 0,
            'lockTime': 0
        }
        result = self.call(url, fsinfo, 'POST')

        msg = 'Failed to create the filesystem.'
        self._assert_result_code(result, msg)

    def _delete_filesystem(self, fs_name):
        """Delete filesystem"""
        url = f'rest/filesystem/{fs_name}'
        result = self.call(url, None, 'DELETE')

        msg = 'Failed to delete the filesystem.'
        self._assert_result_code(result, msg)

    def _get_filesystem(self, fs_name):
        """Get filesystem """
        url = f'rest/filesystem/{fs_name}'
        result = self.call(url, None, 'GET')

        msg = 'Failed to get the filesystem.'
        self._assert_result_code(result, msg)

        return result['data']

    def _create_filesystem_dir(self, share_path):
        url = 'rest/fileDir'
        slash = share_path.index(r'/', 1)
        dir_info = {
            'path': share_path[0: slash],
            'dirName': share_path[slash + 1:],
        }
        result = self.call(url, dir_info, 'POST')

        msg = 'Failed to create the filesystem directory.'
        self._assert_result_code(result, msg)

    def _delete_filesystem_dir(self, share_path):
        slash = share_path.index(r'/', 1)
        url = f'rest/fileDir?path={share_path[0: slash]}' \
              f'&dirName={share_path[slash + 1:]}'

        result = self.call(url, None, 'DELETE')

        msg = 'Failed to delete the filesystem directory.'
        self._assert_result_code(result, msg)

    def _allow_access_rest(self, share_path, access_to,
                           access_level, share_proto):
        """Allow access to the share."""
        if share_proto == 'NFS':
            self._allow_nfs_access_rest(share_path, access_to, access_level)
        elif share_proto == 'CIFS':
            self._allow_cifs_access_rest(share_path, access_to, access_level)
        else:
            raise exception.InvalidInput(
                reason=(_('Invalid Nas protocol: %s.') % share_proto))

    def _allow_nfs_access_rest(self, share_path, access_to, access_level):
        url = 'rest/nfsShareClient'
        access = {
            'path': share_path,
            'client': access_to,
            'authority': access_level,
        }
        result = self.call(url, access, 'POST')

        msg = 'Failed to allow access to the NFS share.'
        self._assert_result_code(result, msg)

    def _allow_cifs_access_rest(self, share_path, access_to, access_level):
        url = 'rest/cifsShareClient'
        ug_type = {
            'localUser': '0',
            'localGroup': '1',
            'adUser': '2',
            'adGroup': '3',
        }

        msg = 'Failed to allow access to the CIFS share.'
        access_info = (f'Access info (access_to: {access_to},'
                       f'access_level: {access_level},'
                       f'path: {share_path}.)')

        def send_rest(rest_access_to, rest_ug_type):
            access = {
                'path': share_path,
                'right': access_level,
                'ugName': rest_access_to,
                'ugType': rest_ug_type,
            }
            result = self.call(url, access, 'POST')
            err_code = result['code']
            if err_code == constants.CODE_SUCCESS:
                return True
            elif err_code != constants.CODE_SOURCE_NOT_EXIST:
                self._assert_result_code(result, msg)
            return False

        if '/' not in access_to:
            # First, try to add local user access
            LOG.debug('Attempting to add local user access. %s', access_info)
            if send_rest(access_to, ug_type['localUser']):
                return
            # Second,If add local user access failed,
            # try to add local group access
            LOG.debug('Failed add local user access,'
                      ' attempting to add local group access. %s', access_info)
            if send_rest(access_to, ug_type['localGroup']):
                return
        else:
            str = access_to.index('/')
            access_to = access_to[str + 1:]
            # First, add domain user access
            LOG.debug('Attempting to add domain user access. %s', access_info)
            if send_rest(access_to, ug_type['adUser']):
                return
            # Second, if add domain user access failed,
            # try to add domain group access.
            LOG.debug('Failed add domain user access, '
                      'attempting to add domain group access. %s', access_info)
            if send_rest(access_to, ug_type['adGroup']):
                return

        raise exception.InvalidShare(reason=msg)

    def _get_access_from_nfs_share(self, path, clientName):
        url = f'rest/nfsShareClient?path={path}&client={clientName}'

        result = self.call(url, None, 'GET')
        msg = 'Failed to get share NFS access.'

        self._assert_result_code(result, msg)
        share_client = None
        if result['data'] is not None:
            share_client = {}
            share_client['path'] = result['data']['path']
            share_client['clientName'] = result['data']['clientName']
            share_client['accessRight'] = result['data']['accessRight']

        return share_client

    def _get_access_from_cifs_share(self, share_path, access_to,
                                    ug_input_type=None):

        ug_type = {
            'localUser': '0',
            'localGroup': '1',
            'adUser': '2',
            'adGroup': '3',
        }

        msg = 'Failed to get share cifs access.'
        access_info = (f'Access info (access_to: {access_to},'
                       f'path: {share_path}.)')

        def send_rest(access_to, ug_type):
            url = f'rest/cifsShareClient?path={share_path}' \
                  f'&ugName={access_to}&ugType={ug_type}'
            result = self.call(url, None, 'GET')
            self._assert_result_code(result, msg)
            return result

        share_client = None
        if ug_input_type is not None:
            ret = send_rest(access_to, ug_input_type)
            if ret['data']:
                share_client = {}
                share_client['path'] = ret['data']['path']
                share_client['ugName'] = ret['data']['ugName']
                share_client['ugType'] = ret['data']['ugType']
                share_client['accessRight'] = ret['data']['accessRight']

            return share_client
        elif '/' not in access_to:
            LOG.debug('Attempting to get local user access. %s', access_info)
            user_ret = send_rest(access_to, ug_type['localUser'])
            if user_ret['code'] == constants.CODE_NOT_FOUND:
                return share_client
            if user_ret['data']:
                share_client = {}
                share_client['path'] = user_ret['data']['path']
                share_client['ugName'] = user_ret['data']['ugName']
                share_client['ugType'] = user_ret['data']['ugType']
                share_client['accessRight'] = user_ret['data']['accessRight']
                return share_client

            LOG.debug('Failed get local user access,'
                      ' attempting to get local group access. %s', access_info)
            group_ret = send_rest(access_to, ug_type['localGroup'])
            if group_ret['data']:
                share_client = {}
                share_client['path'] = group_ret['data']['path']
                share_client['ugName'] = group_ret['data']['ugName']
                share_client['ugType'] = group_ret['data']['ugType']
                share_client['accessRight'] = group_ret['data']['accessRight']
                return share_client
        else:
            str = access_to.index('/')
            access_to = access_to[str + 1:]
            LOG.debug('Attempting to get domain user access. %s', access_info)
            aduser_ret = send_rest(access_to, ug_type['adUser'])
            if aduser_ret['code'] == constants.CODE_NOT_FOUND:
                return share_client
            if aduser_ret['data']:
                share_client = {}
                share_client['path'] = aduser_ret['data']['path']
                share_client['ugName'] = aduser_ret['data']['ugName']
                share_client['ugType'] = aduser_ret['data']['ugType']
                share_client['accessRight'] = \
                    aduser_ret['data']['accessRight']
                return share_client

                LOG.debug('Failed get domain user access,'
                          ' attempting to get domain group access. %s',
                          access_info)
            adgroup_ret = send_rest(access_to, ug_type['adGroup'])
            if adgroup_ret['data']:
                share_client = {}
                share_client['path'] = adgroup_ret['data']['path']
                share_client['ugName'] = adgroup_ret['data']['ugName']
                share_client['ugType'] = adgroup_ret['data']['ugType']
                share_client['accessRight'] = \
                    adgroup_ret['data']['accessRight']
                return share_client

        return share_client

    def _get_all_nfs_access_rest(self, share_path):
        url = f'rest/allNfsShareClient?path={share_path}'

        result = self.call(url, None, 'GET')

        msg = 'Get all nfs access error.'
        self._assert_result_code(result, msg)
        access_list = []
        if result['data'] is None:
            pass
        else:
            for item in result.get('data', []):
                access = {}
                access['share_path'] = item['path']
                access['access_to'] = item['clientName']
                access['access_level'] = item['accessRight']
                access_list.append(access)

        return access_list

    def _get_all_cifs_access_rest(self, share_path):
        url = f'rest/allCifsShareClient?path={share_path}'

        result = self.call(url, None, 'GET')

        msg = 'Get all cifs access error.'
        self._assert_result_code(result, msg)
        access_list = []
        for item in result.get('data', []):
            access = {}
            access['share_path'] = item['path']
            access['access_to'] = item['ugName']
            access['ugType'] = item['ugType']
            access['access_level'] = item['accessRight']
            access_list.append(access)

        return access_list

    def _change_nfs_access_rest(self, share_path, access_to, access_level):
        url = 'rest/nfsShareClient'
        access_info = {
            'path': share_path,
            'oldNfsClientName': access_to,
            'clientName': '',
            'accessRight': access_level,
            'allSquash': '',
            'rootSquash': '',
            'secure': '',
            'anonuid': '',
            'anongid': '',
        }
        result = self.call(url, access_info, 'PUT')

        msg = 'Update nfs acess error.'
        self._assert_result_code(result, msg)

    def _change_cifs_access_rest(self, share_path, access_to,
                                 access_level, ug_type):
        url = 'rest/cifsShareClient'
        if '/' in access_to:
            str = access_to.index('/')
            access_to = access_to[str + 1:]
        access_info = {
            'path': share_path,
            'right': access_level,
            'ugName': access_to,
            'ugType': ug_type,
        }

        result = self.call(url, access_info, 'PUT')
        msg = 'Update cifs access error.'

        self._assert_result_code(result, msg)

    def _delete_nfs_access_rest(self, share_path, access_to):
        url = f'rest/nfsShareClient?path={share_path}&client={access_to}'

        result = self.call(url, None, 'DELETE')
        msg = 'Delete nfs access error.'

        self._assert_result_code(result, msg)

    def _delete_cifs_access_rest(self, share_path, access_to, ug_type):
        url = f'rest/cifsShareClient?path={share_path}&ugName={access_to}' \
              f'&ugType={ug_type}'

        result = self.call(url, None, 'DELETE')
        msg = 'Delete cifs access error.'

        self._assert_result_code(result, msg)

    def _get_nfs_service_status(self):
        url = 'rest/nfsService'
        result = self.call(url, None, 'GET')

        msg = 'Get NFS service stauts error.'
        self._assert_result_code(result, msg)

        nfs_service = {}

        nfs_service['serviceStatus'] = result['data']['serviceStatus']
        nfs_service['nfs3Status'] = result['data']['nfs3Status']
        nfs_service['nfs4Status'] = result['data']['nfs4Status']

        return nfs_service

    def _start_nfs_service(self):
        url = 'rest/nfsService'
        nfs_service_info = {
            "openStatus": "1",
        }

        result = self.call(url, nfs_service_info, 'PUT')

        self._assert_result_code(result, 'Start NFS service error.')

    def _config_nfs_service(self):
        url = 'rest/nfsConfig'
        config_nfs = {
            'configNfs3': "yes",
            'configNfs4': "yes",
        }

        result = self.call(url, config_nfs, 'PUT')

        self._assert_result_code(result, 'Config NFS service error.')

    def _get_cifs_service_status(self):
        url = 'rest/cifsService'
        result = self.call(url, None, 'GET')

        msg = 'Get CIFS service status error.'
        self._assert_result_code(result, msg)

        return result['data']

    def _start_cifs_service(self):
        url = 'rest/cifsService'
        cifs_service_info = {
            'openStatus': '1',
        }

        result = self.call(url, cifs_service_info, 'PUT')

        self._assert_result_code(result, 'Start CIFS service error.')

    def _config_cifs_service(self):
        url = 'rest/cifsConfig'
        """config user mode"""
        config_cifs = {
            'workName': 'manila',
            'description': '',
            'access_way': 'user',
            'isCache': 'no',
            'adsName': '',
            'adsIP': '',
            'adsUSER': '',
            'adsPASSWD': '',
            'allowList': [],
            'denyList': [],
        }

        result = self.call(url, config_cifs, 'PUT')

        self._assert_result_code(result, 'Config CIFS service error.')

    def _get_all_pool(self):
        url = 'rest/storagepool'

        result = self.call(url, None, 'GET')

        msg = 'Query pool info error.'
        self._assert_result_code(result, msg)

        return result

    def _query_user(self, user_name):
        url = f'rest/user/{user_name}'

        result = self.call(url, None, 'GET')

        msg = 'Query user error.'
        self._assert_result_code(result, msg)
        return result['data']

    def _add_localuser(self, user_name, user_passwd, group_name):
        url = 'rest/localUser'
        user_info = {
            'userName': user_name,
            'mgGroup': group_name,
            'userPasswd': user_passwd,
            'unusedGroup': [],
        }
        result = self.call(url, user_info, 'POST')
        msg = 'add localuser error.'
        self._assert_result_code(result, msg)

    def _query_group(self, group_name):
        url = f'rest/group/{group_name}'

        result = self.call(url, None, 'GET')
        msg = 'Query group error.'
        self._assert_result_code(result, msg)
        return result['data']

    def _add_localgroup(self, group_name):
        url = 'rest/localGroup'
        group_info = {
            'groupName': group_name,
        }
        result = self.call(url, group_info, 'POST')

        msg = 'add localgroup error.'
        self._assert_result_code(result, msg)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""The resource_locks API controller module.

This module handles the following requests:
GET /resource-locks
GET /resource-locks/{lock_id}
POST /resource-locks
PUT /resource-locks/{lock_id}
DELETE /resource-locks/{lock_id}
"""

from http import client as http_client

from oslo_utils import timeutils
from oslo_utils import uuidutils
import webob
from webob import exc

from manila.api import common
from manila.api.openstack import wsgi
from manila.api.schemas import resource_locks as schema
from manila.api import validation
from manila.api.views import resource_locks as resource_locks_view
from manila.common import constants
from manila import exception
from manila.i18n import _
from manila.lock import api as resource_locks
from manila import utils

RESOURCE_LOCKS_MIN_API_VERSION = '2.81'


@validation.validated
class ResourceLocksController(wsgi.Controller):
    """The Resource Locks API controller for the OpenStack API."""

    _view_builder_class = resource_locks_view.ViewBuilder
    resource_name = 'resource_lock'

    def _check_body(self, body, lock_to_update=None):
        if 'resource_lock' not in body:
            raise exc.HTTPBadRequest(
                explanation="Malformed request body.")
        lock_data = body['resource_lock']
        resource_type = (
            lock_to_update['resource_type']
            if lock_to_update
            else lock_data.get('resource_type', constants.SHARE_RESOURCE_TYPE)
        )
        resource_id = lock_data.get('resource_id') or ''
        resource_action = (lock_data.get('resource_action') or
                           constants.RESOURCE_ACTION_DELETE)
        lock_reason = lock_data.get('lock_reason') or ''

        if len(lock_reason) > 1023:
            msg = _("'lock_reason' can contain a maximum of 1023 characters.")
            raise exc.HTTPBadRequest(explanation=msg)
        if resource_type not in constants.RESOURCE_LOCK_RESOURCE_TYPES:
            msg = _("'resource_type' is required and must be one "
                    "of %(resource_types)s") % {
                'resource_types': constants.RESOURCE_LOCK_RESOURCE_TYPES
            }
            raise exc.HTTPBadRequest(explanation=msg)
        resource_type_lock_actions = (
            constants.RESOURCE_LOCK_ACTIONS_MAPPING[resource_type])
        if resource_action not in resource_type_lock_actions:
            msg = _("'resource_action' can only be one of %(actions)s" %
                    {'actions': resource_type_lock_actions})
            raise exc.HTTPBadRequest(explanation=msg)

        if lock_to_update:
            if set(lock_data.keys()) - {'resource_action', 'lock_reason'}:
                msg = _("Only 'resource_action' and 'lock_reason' "
                        "can be updated.")
                raise exc.HTTPBadRequest(explanation=msg)
        else:
            if not uuidutils.is_uuid_like(resource_id):
                msg = _("Resource ID is required and must be in uuid format.")
                raise exc.HTTPBadRequest(explanation=msg)

    def __init__(self):
        self.resource_locks_api = resource_locks.API()
        super(ResourceLocksController, self).__init__()

    @wsgi.Controller.api_version(RESOURCE_LOCKS_MIN_API_VERSION)
    @wsgi.Controller.authorize('get_all')
    @validation.request_query_schema(schema.index_request_query)
    @validation.response_body_schema(schema.index_response_body)
    def index(self, req):
        """Returns a list of locks, transformed through view builder."""
        context = req.environ['manila.context']
        filters = req.params.copy()

        params = common.get_pagination_params(req)
        limit, offset = [params.pop('limit', None), params.pop('offset', None)]
        sort_key, sort_dir = common.get_sort_params(filters)
        for key in ('limit', 'offset'):
            filters.pop(key, None)

        show_count = utils.get_bool_from_api_params(
            'with_count', {'with_count': filters.pop('with_count', False)})

        for time_comparison_filter in ['created_since', 'created_before']:
            if time_comparison_filter in filters:
                time_str = filters.get(time_comparison_filter)
                try:
                    parsed_time = timeutils.parse_isotime(time_str)
                    filters[time_comparison_filter] = parsed_time
                except ValueError:
                    msg = _('Invalid value specified for the query '
                            'key: %s') % time_comparison_filter
                    raise exc.HTTPBadRequest(explanation=msg)

        locks, count = self.resource_locks_api.get_all(context,
                                                       search_opts=filters,
                                                       limit=limit,
                                                       offset=offset,
                                                       sort_key=sort_key,
                                                       sort_dir=sort_dir,
                                                       show_count=show_count)

        return self._view_builder.index(req,
                                        locks,
                                        count=count)

    @wsgi.Controller.api_version(RESOURCE_LOCKS_MIN_API_VERSION)
    @wsgi.Controller.authorize('get')
    @validation.request_query_schema(schema.show_request_query)
    @validation.response_body_schema(schema.show_response_body)
    def show(self, req, id):
        """Return an existing resource lock by ID."""
        context = req.environ['manila.context']
        try:
            resource_lock = self.resource_locks_api.get(context, id)
        except exception.ResourceLockNotFound as error:
            raise exc.HTTPNotFound(explanation=error.msg)
        return self._view_builder.detail(req, resource_lock)

    @wsgi.Controller.api_version(RESOURCE_LOCKS_MIN_API_VERSION)
    @wsgi.Controller.authorize
    @wsgi.action("delete")
    @validation.response_body_schema(schema.delete_response_body)
    def delete(self, req, id):
        """Delete an existing resource lock."""
        context = req.environ['manila.context']
        try:
            self.resource_locks_api.delete(context, id)
        except exception.ResourceLockNotFound as error:
            raise exc.HTTPNotFound(explanation=error.msg)
        return webob.Response(status_int=http_client.NO_CONTENT)

    @wsgi.Controller.api_version(RESOURCE_LOCKS_MIN_API_VERSION)
    @wsgi.Controller.authorize
    @validation.request_body_schema(schema.create_request_body)
    @validation.response_body_schema(schema.create_response_body)
    def create(self, req, body):
        """Create a resource lock."""
        context = req.environ['manila.context']
        self._check_body(body)
        lock_data = body['resource_lock']
        try:
            resource_lock = self.resource_locks_api.create(
                context,
                resource_id=lock_data['resource_id'],
                resource_type=lock_data['resource_type'],
                resource_action=(lock_data.get('resource_action') or
                                 constants.RESOURCE_ACTION_DELETE),
                lock_reason=lock_data.get('lock_reason')
            )
        except exception.NotFound:
            raise exc.HTTPBadRequest(
                explanation="No such resource found.")
        except exception.InvalidInput as error:
            raise exc.HTTPConflict(explanation=error.msg)
        except exception.ResourceVisibilityLockExists:
            raise exc.HTTPConflict(
                "Resource's visibility is already locked by other user.")
        return self._view_builder.detail(req, resource_lock)

    @wsgi.Controller.api_version(RESOURCE_LOCKS_MIN_API_VERSION)
    @wsgi.Controller.authorize
    @validation.request_body_schema(schema.update_request_body)
    @validation.response_body_schema(schema.update_response_body)
    def update(self, req, id, body):
        """Update an existing resource lock."""
        context = req.environ['manila.context']
        try:
            resource_lock = self.resource_locks_api.get(context, id)
        except exception.NotFound as e:
            raise exc.HTTPNotFound(explanation=e.msg)

        self._check_body(body, lock_to_update=resource_lock)
        lock_data = body['resource_lock']
        try:
            resource_lock = self.resource_locks_api.update(
                context,
                resource_lock,
                lock_data,
            )
        except exception.InvalidInput as e:
            raise exc.HTTPBadRequest(explanation=e.msg)
        return self._view_builder.detail(req, resource_lock)


def create_resource():
    return wsgi.Resource(ResourceLocksController())



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import ddt
from oslo_config import cfg

from manila.common import constants
from manila import context
from manila import exception
from manila.lock import api as lock_api
from manila import policy
from manila import test
from manila.tests import utils as test_utils
from manila import utils

CONF = cfg.CONF


@ddt.ddt
class ResourceLockApiTest(test.TestCase):

    def setUp(self):
        super(ResourceLockApiTest, self).setUp()
        self.lock_api = lock_api.API()
        self.mock_object(self.lock_api, 'db')
        self.ctxt = context.RequestContext('fakeuser',
                                           'fakeproject',
                                           is_admin=False)
        self.mock_object(policy, 'check_policy')

    @ddt.data(
        test_utils.annotated(
            'admin_context',
            (context.RequestContext('fake', 'fake', is_admin=True), 'admin'),
        ),
        test_utils.annotated(
            'admin_also_service_context',
            (context.RequestContext('fake', 'fake', service_roles=['service'],
                                    is_admin=True), 'service'),
        ),
        test_utils.annotated(
            'service_context',
            (context.RequestContext('fake', 'fake', service_roles=['service'],
                                    is_admin=False), 'service'),
        ),
        test_utils.annotated(
            'user_context',
            (context.RequestContext('fake', 'fake', is_admin=False), 'user')
        ),
    )
    @ddt.unpack
    def test__get_lock_context(self, ctxt, expected_lock_context):
        result = self.lock_api._get_lock_context(ctxt)

        self.assertEqual(expected_lock_context, result['lock_context'])
        self.assertEqual(('fake', 'fake'),
                         (result['user_id'], result['project_id']))

    @ddt.data(
        test_utils.annotated(
            'user_manipulating_admin_lock',
            (context.RequestContext('fake', 'fake', is_admin=False), 'admin'),
        ),
        test_utils.annotated(
            'user_manipulating_service_lock',
            (context.RequestContext('fake', 'fake', is_admin=False),
             'service'),
        ),
        test_utils.annotated(
            'service_manipulating_admin_lock',
            (context.RequestContext('fake', 'fake', is_admin=False,
                                    service_roles=['service']), 'admin'),
        ),
    )
    @ddt.unpack
    def test__check_allow_lock_manipulation_not_allowed(self, ctxt, lock_ctxt):
        self.assertRaises(exception.NotAuthorized,
                          self.lock_api._check_allow_lock_manipulation,
                          ctxt, {'lock_context': lock_ctxt})

    @ddt.data(
        test_utils.annotated(
            'user_manipulating_user_lock',
            (context.RequestContext('fake', 'fake', is_admin=False), 'user'),
        ),
        test_utils.annotated(
            'service_manipulating_service_lock',
            (context.RequestContext(
                'fake', 'fake', is_admin=False, service_roles=['service']),
             'service'),
        ),
        test_utils.annotated(
            'service_manipulating_user_lock',
            (context.RequestContext(
                'fake', 'fake', is_admin=False, service_roles=['service']),
             'user'),
        ),
        test_utils.annotated(
            'admin_manipulating_service_lock',
            (context.RequestContext('fake', 'fake', is_admin=True), 'service'),
        ),
        test_utils.annotated(
            'admin_manipulating_user_lock',
            (context.RequestContext('fake', 'fake', is_admin=True), 'user'),
        ),
    )
    @ddt.unpack
    def test__check_allow_lock_manipulation_allowed(self, ctxt, lock_ctxt):

        result = self.lock_api._check_allow_lock_manipulation(
            ctxt,
            {'lock_context': lock_ctxt}
        )
        self.assertIsNone(result)

    @ddt.data(
        test_utils.annotated(
            'service_manipulating_user_lock',
            (context.RequestContext(
                'fake', 'fake', is_admin=False,
                service_roles=['service']),
             'user',
             'user_b'),
        ),
        test_utils.annotated(
            'admin_manipulating_user_lock',
            (context.RequestContext('fake', 'fake', is_admin=True),
             'admin',
             'user_a'),
        ),
        test_utils.annotated(
            'user_manipulating_locks_they_own',
            (context.RequestContext('user_a', 'fake', is_admin=False),
             'user',
             'user_a'),
        ),
        test_utils.annotated(
            'user_manipulating_other_users_lock',
            (context.RequestContext('user_a', 'fake', is_admin=False),
             'user',
             'user_b'),
        ),
    )
    @ddt.unpack
    def test_access_is_restricted(self, ctxt, lock_ctxt, lock_user):
        resource_lock = {
            'user_id': lock_user,
            'lock_context': lock_ctxt
        }
        is_restricted = (
            (not ctxt.is_admin and not ctxt.is_service)
            and lock_user != ctxt.user_id)
        expected_mock_policy = {}
        if is_restricted:
            expected_mock_policy['side_effect'] = exception.NotAuthorized
        self.mock_object(self.lock_api, '_check_allow_lock_manipulation')
        self.mock_object(policy, 'check_policy',
                         mock.Mock(**expected_mock_policy))

        result = self.lock_api.access_is_restricted(
            ctxt,
            resource_lock
        )
        self.assertEqual(is_restricted, result)

    def test_access_is_restricted_not_authorized(self):
        resource_lock = {
            'user_id': 'fakeuserid',
            'lock_context': 'user'
        }
        ctxt = context.RequestContext('fake', 'fake')
        self.mock_object(self.lock_api, '_check_allow_lock_manipulation',
                         mock.Mock(side_effect=exception.NotAuthorized()))

        result = self.lock_api.access_is_restricted(
            ctxt,
            resource_lock
        )
        self.assertTrue(result)

    def test_get_all_all_projects_ignored(self):
        self.mock_object(policy, 'check_policy', mock.Mock(return_value=False))
        self.mock_object(self.lock_api.db, 'resource_lock_get_all',
                         mock.Mock(return_value=('list of locks', None)))

        locks, count = self.lock_api.get_all(
            self.ctxt,
            search_opts={
                'all_projects': True,
                'project_id': '5dca5323e33b49fca4a5b261c72e612c',
            })
        self.lock_api.db.resource_lock_get_all.assert_called_once_with(
            utils.IsAMatcher(context.RequestContext),
            filters={},
            limit=None,
            offset=None,
            sort_key='created_at',
            sort_dir='desc',
            show_count=False,
        )
        self.assertEqual(('list of locks', None), (locks, count))

    def test_get_all_with_filters(self):
        self.mock_object(self.lock_api.db, 'resource_lock_get_all',
                         mock.Mock(return_value=('list of locks', 4)))
        search_opts = {
            'all_projects': True,
            'project_id': '5dca5323e33b49fca4a5b261c72e612c',
            'resource_type': 'snapshot',
        }
        locks = self.lock_api.get_all(
            self.ctxt,
            limit=3,
            offset=3,
            search_opts=search_opts,
            show_count=True
        )
        self.lock_api.db.resource_lock_get_all.assert_called_once_with(
            utils.IsAMatcher(context.RequestContext),
            filters=search_opts,
            limit=3,
            offset=3,
            sort_key='created_at',
            sort_dir='desc',
            show_count=True,
        )
        self.assertEqual('list of locks', locks[0])
        self.assertEqual(4, locks[1])

    def test_create_lock_resource_not_owned_by_user(self):
        self.mock_object(
            policy,
            'check_policy',
            mock.Mock(side_effect=exception.PolicyNotAuthorized(
                action="resource_lock:create")),
        )

        self.assertRaises(exception.PolicyNotAuthorized,
                          self.lock_api.create,
                          self.ctxt,
                          resource_id='19529cea-0471-4972-adaa-fee8694b7538',
                          resource_type='share',
                          resource_action='delete')
        self.lock_api.db.share_get.assert_called_once_with(
            utils.IsAMatcher(context.RequestContext),
            '19529cea-0471-4972-adaa-fee8694b7538',
        )
        self.lock_api.db.resource_lock_create.assert_not_called()

    @ddt.data(constants.STATUS_DELETING,
              constants.STATUS_ERROR_DELETING,
              constants.STATUS_UNMANAGING,
              constants.STATUS_MANAGE_ERROR_UNMANAGING,
              constants.STATUS_UNMANAGE_ERROR,
              constants.STATUS_UNMANAGED,
              constants.STATUS_DELETED)
    def test_create_lock_invalid_resource_status(self, status):
        self.mock_object(self.lock_api.db, 'resource_lock_create',
                         mock.Mock(return_value='created_obj'))
        self.mock_object(self.lock_api.db, 'share_get',
                         mock.Mock(return_value={'status': status}))

        self.assertRaises(exception.InvalidInput,
                          self.lock_api.create,
                          self.ctxt,
                          resource_id='7dab6090-1dfd-4829-bbaf-602fcd1c8248',
                          resource_action='delete',
                          resource_type='share')

        self.lock_api.db.resource_lock_create.assert_not_called()

    def test_create_lock_invalid_resource_soft_deleted(self):
        self.mock_object(self.lock_api.db, 'resource_lock_create',
                         mock.Mock(return_value='created_obj'))
        self.mock_object(self.lock_api.db, 'share_get',
                         mock.Mock(return_value={'is_soft_deleted': True}))

        self.assertRaises(exception.InvalidInput,
                          self.lock_api.create,
                          self.ctxt,
                          resource_id='0bbf0b62-cb29-4218-920b-3f62faa99ff8',
                          resource_action='delete',
                          resource_type='share')

        self.lock_api.db.resource_lock_create.assert_not_called()

    def test_create_lock(self):
        self.mock_object(self.lock_api.db, 'resource_lock_create',
                         mock.Mock(return_value='created_obj'))
        mock_share = {
            'id': 'cacac01c-853d-47f3-afcb-da4484bd09a5',
            'status': constants.STATUS_AVAILABLE,
            'is_soft_deleted': False,
        }
        self.mock_object(self.lock_api.db, 'share_get',
                         mock.Mock(return_value=mock_share))

        result = self.lock_api.create(
            self.ctxt,
            resource_id='cacac01c-853d-47f3-afcb-da4484bd09a5',
            resource_action='delete',
            resource_type='share',
        )

        self.assertEqual('created_obj', result)
        db_create_arg = self.lock_api.db.resource_lock_create.call_args[0][1]
        expected_create_arg = {
            'resource_id': 'cacac01c-853d-47f3-afcb-da4484bd09a5',
            'resource_action': 'delete',
            'user_id': 'fakeuser',
            'project_id': 'fakeproject',
            'lock_context': 'user',
            'lock_reason': None,
            'resource_type': constants.SHARE_RESOURCE_TYPE

        }
        self.assertEqual(expected_create_arg, db_create_arg)

    def test_create_access_show_lock(self):
        self.mock_object(self.lock_api.db, 'resource_lock_create',
                         mock.Mock(return_value='created_obj'))
        mock_access = {
            'id': 'cacac01c-853d-47f3-afcb-da4484bd09a5',
            'state': constants.STATUS_ACTIVE,
        }
        self.mock_object(self.lock_api.db, 'access_get',
                         mock.Mock(return_value=mock_access))
        self.mock_object(self.lock_api.db, 'resource_lock_get_all',
                         mock.Mock(return_value=['', 0]))
        self.mock_object(self.ctxt, 'elevated',
                         mock.Mock(return_value=self.ctxt))

        result = self.lock_api.create(
            self.ctxt,
            resource_id='cacac01c-853d-47f3-afcb-da4484bd09a5',
            resource_action=constants.RESOURCE_ACTION_SHOW,
            resource_type=constants.SHARE_ACCESS_RESOURCE_TYPE,
        )

        self.assertEqual('created_obj', result)
        db_create_arg = self.lock_api.db.resource_lock_create.call_args[0][1]
        resource_id = 'cacac01c-853d-47f3-afcb-da4484bd09a5'
        expected_create_arg = {
            'resource_id': resource_id,
            'resource_action': constants.RESOURCE_ACTION_SHOW,
            'user_id': 'fakeuser',
            'project_id': 'fakeproject',
            'lock_context': 'user',
            'lock_reason': None,
            'resource_type': constants.SHARE_ACCESS_RESOURCE_TYPE

        }
        self.assertEqual(expected_create_arg, db_create_arg)
        filters = {
            'resource_id': resource_id,
            'resource_action': constants.RESOURCE_ACTION_SHOW,
            'all_projects': True
        }
        self.lock_api.db.resource_lock_get_all.assert_called_once_with(
            self.ctxt, filters=filters)

    def test_create_visibility_lock_lock_exists(self):
        self.mock_object(self.lock_api.db, 'resource_lock_create',
                         mock.Mock(return_value='created_obj'))
        self.mock_object(self.lock_api.db, 'resource_lock_get_all',
                         mock.Mock(return_value=['visibility_lock', 1]))
        self.mock_object(self.ctxt, 'elevated',
                         mock.Mock(return_value=self.ctxt))

        self.assertRaises(
            exception.ResourceVisibilityLockExists,
            self.lock_api.create,
            self.ctxt,
            resource_id='cacac01c-853d-47f3-afcb-da4484bd09a5',
            resource_action=constants.RESOURCE_ACTION_SHOW,
            resource_type=constants.SHARE_ACCESS_RESOURCE_TYPE,
        )

        resource_id = 'cacac01c-853d-47f3-afcb-da4484bd09a5'
        filters = {
            'resource_id': resource_id,
            'resource_action': constants.RESOURCE_ACTION_SHOW,
            'all_projects': True
        }
        self.lock_api.db.resource_lock_get_all.assert_called_once_with(
            self.ctxt, filters=filters)

    @ddt.data(True, False)
    def test_update_lock_resource_not_allowed_with_policy_failure(
            self, policy_fails):
        lock = {'id': 'd767d3cd-1187-404a-a91f-8b172e0e768e'}
        if policy_fails:
            self.mock_object(
                policy,
                'check_policy',
                mock.Mock(
                    side_effect=exception.PolicyNotAuthorized(
                        action='resource_lock:update'),
                ),
            )
        self.mock_object(
            self.lock_api,
            '_check_allow_lock_manipulation',
            mock.Mock(
                side_effect=exception.NotAuthorized
            ),
        )

        self.assertRaises(exception.NotAuthorized,
                          self.lock_api.update,
                          self.ctxt,
                          lock,
                          {'foo': 'bar'})

    @ddt.data(constants.STATUS_DELETING,
              constants.STATUS_ERROR_DELETING,
              constants.STATUS_UNMANAGING,
              constants.STATUS_MANAGE_ERROR_UNMANAGING,
              constants.STATUS_UNMANAGE_ERROR,
              constants.STATUS_UNMANAGED,
              constants.STATUS_DELETED)
    def test_update_invalid_resource_status(self, status):
        lock = {
            'id': 'd767d3cd-1187-404a-a91f-8b172e0e768e',
            'resource_id': '266cf54f-f9cf-4d6c-94f3-7b67f00e0465',
            'resource_action': 'something',
            'resource_type': 'share',
        }
        self.mock_object(self.lock_api, '_check_allow_lock_manipulation')
        self.mock_object(self.lock_api.db,
                         'share_get',
                         mock.Mock(return_value={'status': status}))

        self.assertRaises(exception.InvalidInput,
                          self.lock_api.update,
                          self.ctxt,
                          lock,
                          {'resource_action': 'delete'})

        self.lock_api.db.resource_lock_update.assert_not_called()

    def test_update(self):
        self.mock_object(self.lock_api, '_check_allow_lock_manipulation')
        self.mock_object(self.lock_api.db, 'resource_lock_update',
                         mock.Mock(return_value='updated_obj'))
        lock = {'id': 'd767d3cd-1187-404a-a91f-8b172e0e768e'}

        result = self.lock_api.update(
            self.ctxt,
            lock,
            {'foo': 'bar'},
        )

        self.assertEqual('updated_obj', result)
        self.lock_api.db.resource_lock_update.assert_called_once_with(
            utils.IsAMatcher(context.RequestContext),
            'd767d3cd-1187-404a-a91f-8b172e0e768e',
            {'foo': 'bar'},
        )

    @ddt.data(True, False)
    def test_delete_not_allowed_with_policy_failure(self, policy_fails):
        self.mock_object(self.lock_api.db, 'resource_lock_get', mock.Mock(
            return_value={'id': 'd767d3cd-1187-404a-a91f-8b172e0e768e'}))
        if policy_fails:
            self.mock_object(
                policy,
                'check_policy',
                mock.Mock(
                    side_effect=exception.PolicyNotAuthorized(
                        action='resource_lock:delete'),
                ),
            )
        self.mock_object(
            self.lock_api,
            '_check_allow_lock_manipulation',
            mock.Mock(
                side_effect=exception.NotAuthorized
            ),
        )

        self.assertRaises(exception.NotAuthorized,
                          self.lock_api.delete,
                          self.ctxt,
                          'd767d3cd-1187-404a-a91f-8b172e0e768e')

        policy.check_policy.assert_called_once_with(
            utils.IsAMatcher(context.RequestContext),
            'resource_lock',
            'delete',
            {'id': 'd767d3cd-1187-404a-a91f-8b172e0e768e'},
        )
        self.assertEqual(not policy_fails,
                         self.lock_api._check_allow_lock_manipulation.called)
        self.lock_api.db.resource_lock_delete.assert_not_called()

    def test_delete(self):
        self.mock_object(self.lock_api.db, 'resource_lock_get', mock.Mock(
            return_value={'id': 'd767d3cd-1187-404a-a91f-8b172e0e768e'}))
        self.mock_object(self.lock_api, '_check_allow_lock_manipulation')

        result = self.lock_api.delete(self.ctxt,
                                      'd767d3cd-1187-404a-a91f-8b172e0e768e')
        self.assertIsNone(result)
        self.lock_api.db.resource_lock_delete.assert_called_once_with(
            utils.IsAMatcher(context.RequestContext),
            'd767d3cd-1187-404a-a91f-8b172e0e768e'
        )

    def test_ensure_context_can_delete_lock_policy_fails(self):
        lock = {'id': 'd767d3cd-1187-404a-a91f-8b172e0e768e'}
        self.mock_object(
            self.lock_api.db, 'resource_lock_get', mock.Mock(return_value=lock)
        )
        self.mock_object(
            policy,
            'check_policy',
            mock.Mock(side_effect=exception.PolicyNotAuthorized(
                action="resource_lock:delete")),
        )

        self.assertRaises(
            exception.NotAuthorized,
            self.lock_api.ensure_context_can_delete_lock,
            self.ctxt,
            'd767d3cd-1187-404a-a91f-8b172e0e768e')

        self.lock_api.db.resource_lock_get.assert_called_once_with(
            self.ctxt, 'd767d3cd-1187-404a-a91f-8b172e0e768e'
        )
        policy.check_policy.assert_called_once_with(
            self.ctxt, 'resource_lock', 'delete', lock)

    def test_ensure_context_can_delete_lock(self):
        lock = {'id': 'd767d3cd-1187-404a-a91f-8b172e0e768e'}
        self.mock_object(
            self.lock_api.db, 'resource_lock_get', mock.Mock(return_value=lock)
        )
        self.mock_object(policy, 'check_policy')
        self.mock_object(self.lock_api, '_check_allow_lock_manipulation')

        self.lock_api.ensure_context_can_delete_lock(
            self.ctxt,
            'd767d3cd-1187-404a-a91f-8b172e0e768e')

        self.lock_api.db.resource_lock_get.assert_called_once_with(
            self.ctxt, 'd767d3cd-1187-404a-a91f-8b172e0e768e'
        )
        policy.check_policy.assert_called_once_with(
            self.ctxt, 'resource_lock', 'delete', lock)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""The Share Backups API."""

import webob
from webob import exc

from manila.api import common
from manila.api.openstack import wsgi
from manila.api.views import share_backups as backup_view
from manila import db
from manila import exception
from manila.i18n import _
from manila import policy
from manila import share


MIN_SUPPORTED_API_VERSION = '2.80'


class ShareBackupController(wsgi.Controller, wsgi.AdminActionsMixin):
    """The Share Backup API controller for the OpenStack API."""

    resource_name = 'share_backup'
    _view_builder_class = backup_view.BackupViewBuilder

    def __init__(self):
        super(ShareBackupController, self).__init__()
        self.share_api = share.API()

    def _update(self, *args, **kwargs):
        db.share_backup_update(*args, **kwargs)

    def _get(self, *args, **kwargs):
        return db.share_backup_get(*args, **kwargs)

    @wsgi.Controller.api_version(MIN_SUPPORTED_API_VERSION, experimental=True)
    def index(self, req):
        """Return a summary list of backups."""
        return self._get_backups(req)

    @wsgi.Controller.api_version(MIN_SUPPORTED_API_VERSION, experimental=True)
    def detail(self, req):
        """Returns a detailed list of backups."""
        return self._get_backups(req, is_detail=True)

    @wsgi.Controller.authorize('get_all')
    def _get_backups(self, req, is_detail=False):
        """Returns list of backups."""
        context = req.environ['manila.context']

        search_opts = {}
        search_opts.update(req.GET)
        params = common.get_pagination_params(req)
        limit, offset = [params.get('limit'), params.get('offset')]

        search_opts.pop('limit', None)
        search_opts.pop('offset', None)
        sort_key, sort_dir = common.get_sort_params(search_opts)
        key_dict = {"name": "display_name",
                    "description": "display_description"}
        for key in key_dict:
            if sort_key == key:
                sort_key = key_dict[key]

        if 'name' in search_opts:
            search_opts['display_name'] = search_opts.pop('name')
        if 'description' in search_opts:
            search_opts['display_description'] = search_opts.pop(
                'description')

        # like filter
        for key, db_key in (('name~', 'display_name~'),
                            ('description~', 'display_description~')):
            if key in search_opts:
                search_opts[db_key] = search_opts.pop(key)

        common.remove_invalid_options(context, search_opts,
                                      self._get_backups_search_options())

        # Read and remove key 'all_tenants' if was provided
        search_opts['project_id'] = context.project_id
        all_tenants = search_opts.pop('all_tenants',
                                      search_opts.pop('all_projects', None))
        if all_tenants:
            allowed_to_list_all_tenants = policy.check_policy(
                context, 'share_backup', 'get_all_project', do_raise=False)
            if allowed_to_list_all_tenants:
                search_opts.pop('project_id')

        share_id = req.params.get('share_id')
        if share_id:
            try:
                self.share_api.get(context, share_id)
                search_opts.update({'share_id': share_id})
            except exception.NotFound:
                msg = _("No share exists with ID %s.")
                raise exc.HTTPBadRequest(explanation=msg % share_id)

        backups = db.share_backups_get_all(context,
                                           filters=search_opts,
                                           limit=limit,
                                           offset=offset,
                                           sort_key=sort_key,
                                           sort_dir=sort_dir)
        if is_detail:
            backups = self._view_builder.detail_list(req, backups)
        else:
            backups = self._view_builder.summary_list(req, backups)

        return backups

    def _get_backups_search_options(self):
        """Return share backup search options allowed by non-admin."""
        return ('display_name', 'status', 'share_id', 'topic', 'display_name~',
                'display_description~', 'display_description')

    @wsgi.Controller.api_version(MIN_SUPPORTED_API_VERSION, experimental=True)
    @wsgi.Controller.authorize('get')
    def show(self, req, id):
        """Return data about the given backup."""
        context = req.environ['manila.context']

        try:
            backup = db.share_backup_get(context, id)
        except exception.ShareBackupNotFound:
            msg = _("No backup exists with ID %s.")
            raise exc.HTTPNotFound(explanation=msg % id)

        return self._view_builder.detail(req, backup)

    @wsgi.Controller.api_version(MIN_SUPPORTED_API_VERSION, experimental=True)
    @wsgi.Controller.authorize
    @wsgi.response(202)
    def create(self, req, body):
        """Add a backup to an existing share."""
        context = req.environ['manila.context']

        if not self.is_valid_body(body, 'share_backup'):
            msg = _("Body does not contain 'share_backup' information.")
            raise exc.HTTPUnprocessableEntity(explanation=msg)

        backup = body.get('share_backup')
        share_id = backup.get('share_id')

        if not share_id:
            msg = _("'share_id' is missing from the request body.")
            raise exc.HTTPBadRequest(explanation=msg)

        try:
            share = self.share_api.get(context, share_id)
        except exception.NotFound:
            msg = _("No share exists with ID %s.")
            raise exc.HTTPBadRequest(explanation=msg % share_id)
        if share.get('is_soft_deleted'):
            msg = _("Backup can not be created for share '%s' "
                    "since it has been soft deleted.") % share_id
            raise exc.HTTPForbidden(explanation=msg)

        try:
            backup = self.share_api.create_share_backup(context, share, backup)
        except (exception.InvalidBackup,
                exception.InvalidShare) as e:
            raise exc.HTTPBadRequest(explanation=e.msg)
        except exception.ShareBusyException as e:
            raise exc.HTTPConflict(explanation=e.msg)

        return self._view_builder.detail(req, backup)

    @wsgi.Controller.api_version(MIN_SUPPORTED_API_VERSION, experimental=True)
    @wsgi.Controller.authorize
    def delete(self, req, id):
        """Delete a backup."""
        context = req.environ['manila.context']

        try:
            backup = db.share_backup_get(context, id)
        except exception.ShareBackupNotFound:
            msg = _("No backup exists with ID %s.")
            raise exc.HTTPNotFound(explanation=msg % id)

        try:
            self.share_api.delete_share_backup(context, backup)
        except exception.InvalidBackup as e:
            raise exc.HTTPBadRequest(explanation=e.msg)

        return webob.Response(status_int=202)

    @wsgi.Controller.api_version(MIN_SUPPORTED_API_VERSION, experimental=True)
    @wsgi.action('restore')
    @wsgi.Controller.authorize
    @wsgi.response(202)
    def restore(self, req, id, body):
        """Restore an existing backup to a share."""
        context = req.environ['manila.context']

        try:
            backup = db.share_backup_get(context, id)
        except exception.ShareBackupNotFound:
            msg = _("No backup exists with ID %s.")
            raise exc.HTTPNotFound(explanation=msg % id)

        try:
            restored = self.share_api.restore_share_backup(context, backup)
        except (exception.InvalidShare,
                exception.InvalidBackup) as e:
            raise exc.HTTPBadRequest(explanation=e.msg)

        retval = self._view_builder.restore_summary(req, restored)
        return retval

    @wsgi.Controller.api_version(MIN_SUPPORTED_API_VERSION, experimental=True)
    @wsgi.Controller.authorize
    @wsgi.response(200)
    def update(self, req, id, body):
        """Update a backup."""
        context = req.environ['manila.context']

        if not self.is_valid_body(body, 'share_backup'):
            msg = _("Body does not contain 'share_backup' information.")
            raise exc.HTTPUnprocessableEntity(explanation=msg)

        try:
            backup = db.share_backup_get(context, id)
        except exception.ShareBackupNotFound:
            msg = _("No backup exists with ID %s.")
            raise exc.HTTPNotFound(explanation=msg % id)

        backup_update = body.get('share_backup')
        update_dict = {}
        if 'name' in backup_update:
            update_dict['display_name'] = backup_update.pop('name')
        if 'description' in backup_update:
            update_dict['display_description'] = (
                backup_update.pop('description'))

        backup = self.share_api.update_share_backup(context, backup,
                                                    update_dict)
        return self._view_builder.detail(req, backup)

    @wsgi.Controller.api_version(MIN_SUPPORTED_API_VERSION, experimental=True)
    @wsgi.action('reset_status')
    def backup_reset_status(self, req, id, body):
        return self._reset_status(req, id, body)


def create_resource():
    return wsgi.Resource(ShareBackupController())



# Copyright 2024 VAST Data Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from abc import ABC
import json
import pprint
import textwrap

import cachetools
from oslo_log import log as logging
from oslo_utils import versionutils
from packaging import version as packaging_version
import requests

from manila import exception
from manila.share.drivers.vastdata import driver_util
import manila.utils as manila_utils

LOG = logging.getLogger(__name__)


class Session(requests.Session):

    def __init__(self, host, username, password, ssl_verify, plugin_version):
        super().__init__()
        self.base_url = f"https://{host.strip('/')}/api"
        self.ssl_verify = ssl_verify
        self.username = username
        self.password = password
        self.headers["Accept"] = "application/json"
        self.headers["Content-Type"] = "application/json"
        self.headers["User-Agent"] = (
            f"manila/v{plugin_version}"
            f" ({requests.utils.default_user_agent()})"
        )
        # will be updated on first request
        self.headers["authorization"] = "Bearer"

        if not ssl_verify:
            import urllib3

            urllib3.disable_warnings()

    def refresh_auth_token(self):
        try:
            resp = super().request(
                "POST",
                f"{self.base_url}/token/",
                verify=self.ssl_verify,
                timeout=5,
                json={"username": self.username, "password": self.password},
            )
            resp.raise_for_status()
            token = resp.json()["access"]
            self.headers["authorization"] = f"Bearer {token}"
        except ConnectionError as e:
            raise exception.VastApiException(
                reason=f"The vms on the designated host {self.base_url} "
                f"cannot be accessed. Please verify the specified endpoint. "
                f"origin error: {e}"
            )

    @manila_utils.retry(retry_param=exception.VastApiRetry, retries=3)
    def request(
            self, verb, api_method, params=None, log_result=True, **kwargs
    ):
        verb = verb.upper()
        api_method = api_method.strip("/")
        url = f"{self.base_url}/{api_method}/"
        log_pref = f"\n[{verb}] {url}"

        if "data" in kwargs:
            kwargs["data"] = json.dumps(kwargs["data"])

        if log_result and (params or kwargs):
            payload = dict(kwargs, params=params)
            formatted_request = textwrap.indent(
                pprint.pformat(payload), prefix="|  "
            )
            LOG.debug(f"{log_pref} >>>:\n{formatted_request}")
        else:
            LOG.debug(f"{log_pref} >>> (request)")

        ret = super().request(
            verb, url, verify=self.ssl_verify, params=params, **kwargs
        )
        if ret.status_code == 403 and "Token is invalid" in ret.text:
            self.refresh_auth_token()
            raise exception.VastApiRetry(reason="Token is invalid or expired.")

        if ret.status_code in (400, 503) and ret.text:
            raise exception.VastApiException(reason=ret.text)

        try:
            ret.raise_for_status()
        except Exception as exc:
            raise exception.VastApiException(reason=str(exc))

        ret = ret.json() if ret.content else {}
        if ret and log_result:
            formatted_response = textwrap.indent(
                pprint.pformat(ret), prefix="|  "
            )
            LOG.debug(f"{log_pref} <<<:\n{formatted_response}")
        else:
            LOG.debug(f"{log_pref} <<< (response)")
        return driver_util.Bunch.from_dict(ret)

    def __getattr__(self, attr):
        if attr.startswith("_"):
            raise AttributeError(attr)

        def func(**params):
            return self.request("get", attr, params=params)

        func.__name__ = attr
        setattr(self, attr, func)
        return func


def requisite(semver: str, operation: str = None):
    """Use this decorator to indicate the minimum required version cluster

     for invoking the API that is being decorated.
    Decorator works in two modes:
    1. When ignore == False and version mismatch detected then
        `OperationNotSupported` exception will be thrown
    2. When ignore == True and version mismatch detected then
        method decorated method execution never happened
    """

    def dec(fn):

        def _args_wrapper(self, *args, **kwargs):

            version = packaging_version.parse(
                self.rest.get_sw_version().replace("-", ".")
            )
            sw_version = f"{version.major}.{version.minor}.{version.micro}"

            if not versionutils.is_compatible(
                    semver, sw_version, same_major=False
            ):
                op = operation or fn.__name__
                raise exception.VastDriverException(
                    f"Operation {op} is not supported"
                    f" on VAST version {sw_version}."
                    f" Required version is {semver}"
                )
            return fn(self, *args, **kwargs)

        return _args_wrapper

    return dec


class VastResource(ABC):
    resource_name = None

    def __init__(self, rest):
        self.rest = rest  # For intercommunication between resources.
        self.session = rest.session

    def list(self, **params):
        """Get list of entries with optional filtering params"""
        return self.session.get(self.resource_name, params=params)

    def create(self, **params):
        """Create new entry with provided params"""
        return self.session.post(self.resource_name, data=params)

    def update(self, entry_id, **params):
        """Update entry by id with provided params"""
        return self.session.patch(
            f"{self.resource_name}/{entry_id}", data=params
        )

    def delete(self, name):
        """Delete entry by name. Skip if entry not found."""
        entry = self.one(name)
        if not entry:
            resource = self.__class__.__name__.lower()
            LOG.warning(
                f"{resource} {name} not found on VAST, skipping delete"
            )
            return
        return self.session.delete(f"{self.resource_name}/{entry.id}")

    def one(self, name):
        """Get single entry by name.

         Raise exception if multiple entries found.
         """
        entries = self.list(name=name)
        if not entries:
            return
        if len(entries) > 1:
            resource = self.__class__.__name__.lower() + "s"
            raise exception.VastDriverException(
                reason=f"Too many {resource} found with name {name}"
            )
        return entries[0]

    def ensure(self, name, **params):
        entry = self.one(name)
        if not entry:
            entry = self.create(name=name, **params)
        return entry


class View(VastResource):
    resource_name = "views"

    def create(self, name, path, policy_id):
        data = dict(
            name=name,
            path=path,
            policy_id=policy_id,
            create_dir=True,
            protocols=["NFS"],
        )
        return super().create(**data)


class ViewPolicy(VastResource):
    resource_name = "viewpolicies"


class CapacityMetrics(VastResource):

    def get(self, metrics, object_type="cluster", time_frame="1m"):
        """Get capacity metrics for the cluster"""
        params = dict(
            prop_list=metrics,
            object_type=object_type, time_frame=time_frame
        )
        ret = self.session.get("monitors/ad_hoc_query", params=params)
        last_sample = ret.data[-1]
        return driver_util.Bunch(
            {
                name.partition(",")[-1]: value
                for name, value in zip(ret.prop_list, last_sample)
            }
        )


class Quota(VastResource):
    resource_name = "quotas"


class VipPool(VastResource):
    resource_name = "vippools"

    def vips(self, pool_name):
        """Get list of ip addresses from vip pool"""
        vippool = self.one(name=pool_name)
        if not vippool:
            raise exception.VastDriverException(
                reason=f"No vip pool found with name {pool_name}"
            )
        vips = driver_util.generate_ip_range(vippool.ip_ranges)
        if not vips:
            raise exception.VastDriverException(
                reason=f"Pool {pool_name} has no available vips"
            )
        return vips


class Snapshots(VastResource):
    resource_name = "snapshots"


class Folders(VastResource):
    resource_name = "folders"

    @requisite(semver="4.7.0")
    def delete(self, path):
        try:
            self.session.delete(
                f"{self.resource_name}/delete_folder/", data=dict(path=path)
            )
        except exception.VastApiException as e:
            exc_msg = str(e)
            if "no such directory" in exc_msg:
                LOG.debug(f"remote directory "
                          f"might have been removed earlier. ({e})")
            elif "trash folder disabled" in exc_msg:
                raise exception.VastDriverException(
                    reason="Trash Folder Access is disabled"
                    " (see Settings/Cluster/Features in VMS)"
                )
            else:
                # unpredictable error
                raise


class RestApi:

    def __init__(self, host, username, password, ssl_verify, plugin_version):
        self.session = Session(
            host=host,
            username=username,
            password=password,
            ssl_verify=ssl_verify,
            plugin_version=plugin_version,
        )
        self.views = View(self)
        self.view_policies = ViewPolicy(self)
        self.capacity_metrics = CapacityMetrics(self)
        self.quotas = Quota(self)
        self.vip_pools = VipPool(self)
        self.snapshots = Snapshots(self)
        self.folders = Folders(self)

        # Refresh auth token to avoid initial "forbidden" status error.
        self.session.refresh_auth_token()

    @cachetools.cached(cache=cachetools.TTLCache(ttl=60 * 60, maxsize=1))
    def get_sw_version(self):
        """Software version of cluster Rest API interacts with"""
        return self.session.versions(status="success")[0].sys_version



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""add mount_point_name to share_instances

Revision ID: 6e32091979e0
Revises: 99d328f0a3d2
Create Date: 2024-01-26 22:08:22.412974
"""
# revision identifiers, used by Alembic.
revision = '6e32091979e0'
down_revision = '99d328f0a3d2'

from alembic import op
from oslo_log import log
import sqlalchemy as sa

LOG = log.getLogger(__name__)
share_instances_table_name = 'share_instances'
column_name = "mount_point_name"


def upgrade():
    try:
        op.add_column(share_instances_table_name, sa.Column(column_name,
                                                            sa.String(255),
                                                            nullable=True))
    except Exception:
        LOG.error("Column mount_point_name not created!")
        raise


def downgrade():
    try:
        op.drop_column(share_instances_table_name, column_name)
    except Exception:
        LOG.error("Column mount_point_name not dropped!")
        raise



# Copyright 2021 Cloudification GmbH.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ddt

from manila import context
from manila.scheduler.filters import host
from manila import test
from manila.tests.scheduler import fakes


fake_host1 = fakes.FakeHostState('host1', {})
fake_host2 = fakes.FakeHostState('host2', {})


@ddt.ddt
class OnlyHostFilterTestCase(test.TestCase):
    """Test case for OnlyHostFilter."""

    def setUp(self):
        super(OnlyHostFilterTestCase, self).setUp()
        self.filter = host.OnlyHostFilter()
        self.user_context = context.RequestContext('user', 'project')
        self.admin_context = context.RequestContext('user', 'project',
                                                    is_admin=True)

    def _make_filter_properties(self, hint):
        return {
            'context': self.admin_context,
            'scheduler_hints': hint,
        }

    @ddt.data((fake_host1, {'scheduler_hints': None}),
              (fake_host1, {'scheduler_hints': {}}),
              (fake_host1,
              {'scheduler_hints': {'only_host': fake_host2.host}}))
    @ddt.unpack
    def test_only_host_filter_user_context(self, host, filter_properties):
        context = {'context': self.user_context}
        filter_properties.update(context)
        self.assertTrue(self.filter.host_passes(host, filter_properties))

    @ddt.data((fake_host1, None, True),
              (fake_host1, {}, True),
              (fake_host1, {'only_host': fake_host1.host}, True),
              (fake_host2, {'only_host': fake_host1.host}, False))
    @ddt.unpack
    def test_only_host_filter_admin_context(self, host, hint, host_passes):
        filter_properties = self._make_filter_properties(hint)
        self.assertEqual(host_passes,
                         self.filter.host_passes(host, filter_properties))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""add_disable_reason_to_service

Revision ID: 99d328f0a3d2
Revises: 2d708a9a3ba9
Create Date: 2023-10-13 10:50:29.311032

"""

# revision identifiers, used by Alembic.
revision = '99d328f0a3d2'
down_revision = '2d708a9a3ba9'

from alembic import op
from oslo_log import log
import sqlalchemy as sa


LOG = log.getLogger(__name__)


def upgrade():
    try:
        op.add_column(
            'services', sa.Column(
                'disabled_reason', sa.String(length=255),
                nullable=True))
    except Exception:
        LOG.error("Column services.disabled_reason not created!")
        raise


def downgrade():
    try:
        op.drop_column('services', 'disabled_reason')
    except Exception:
        LOG.error("Column services.disabled_reason not dropped!")
        raise



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""multiple share server subnets

Revision ID: a87e0fb17dee
Revises: 1946cb97bb8d
Create Date: 2022-01-14 06:12:27.596130

"""

# revision identifiers, used by Alembic.
revision = 'a87e0fb17dee'
down_revision = '1946cb97bb8d'


from alembic import op
from oslo_log import log
import sqlalchemy as sa

from manila.db.migrations import utils


SHARE_SERVERS_TABLE = 'share_servers'
SHARE_SERVER_SUBNET_MAP_TABLE = 'share_server_share_network_subnet_mappings'
NETWORK_ALLOCATIONS_TABLE = 'network_allocations'
LOG = log.getLogger(__name__)


def upgrade():

    # Create mappings table.
    context = op.get_context()
    mysql_dl = context.bind.dialect.name == 'mysql'
    datetime_type = (sa.dialects.mysql.DATETIME(fsp=6)
                     if mysql_dl else sa.DateTime)
    try:
        share_server_fk_name = "fk_ss_sns_m_share_server_id_share_servers"
        share_network_subnet_fk_name = (
            "fk_ss_sns_m_share_network_subnet_id_share_network_subnets")
        server_subnet_mappings_table = op.create_table(
            SHARE_SERVER_SUBNET_MAP_TABLE,
            sa.Column('id', sa.Integer, primary_key=True, nullable=False),
            sa.Column('created_at', datetime_type),
            sa.Column('updated_at', datetime_type),
            sa.Column('deleted_at', datetime_type),
            sa.Column('deleted', sa.Integer, default=0),
            sa.Column(
                'share_server_id', sa.String(length=36),
                sa.ForeignKey('share_servers.id', name=share_server_fk_name),
                nullable=False),
            sa.Column(
                'share_network_subnet_id', sa.String(length=36),
                sa.ForeignKey('share_network_subnets.id',
                              name=share_network_subnet_fk_name),
                nullable=False),
            mysql_engine='InnoDB',
            mysql_charset='utf8')
    except Exception:
        LOG.error('Table %s could not be created.',
                  SHARE_SERVER_SUBNET_MAP_TABLE)
        raise

    # Populate the mappings table from the share servers table.
    try:
        connection = op.get_bind()
        share_servers_table = utils.load_table(SHARE_SERVERS_TABLE, connection)
        server_subnet_mappings = []
        for server in connection.execute(share_servers_table.select()):
            if server.share_network_subnet_id:
                server_subnet_mappings.append({
                    'created_at': server.created_at,
                    'updated_at': server.updated_at,
                    'deleted_at': server.deleted_at,
                    'deleted': 0 if server.deleted == 'False' else 1,
                    'share_server_id': server.id,
                    'share_network_subnet_id': server.share_network_subnet_id,
                })
        op.bulk_insert(server_subnet_mappings_table, server_subnet_mappings)
    except Exception:
        LOG.error('Table %s could not be populated from the %s table.',
                  SHARE_SERVER_SUBNET_MAP_TABLE, SHARE_SERVERS_TABLE)
        raise

    # add subnet id column to the allocations table.
    try:
        network_allocation_fk_name = (
            "fk_network_allocation_subnet_id_share_network_subnets")
        op.add_column(
            NETWORK_ALLOCATIONS_TABLE,
            sa.Column('share_network_subnet_id', sa.String(length=36),
                      sa.ForeignKey('share_network_subnets.id',
                                    name=network_allocation_fk_name))
        )
    except Exception:
        LOG.error("Could not add ForeignKey column 'share_network_subnet_id'"
                  "to table %s.", NETWORK_ALLOCATIONS_TABLE)
        raise

    # populate the allocation with its subnet id using the share server.
    network_allocation_table = utils.load_table(NETWORK_ALLOCATIONS_TABLE,
                                                connection)
    for alloc in connection.execute(network_allocation_table.select()):
        # admin allocations should not contain subnet id.
        if alloc._mapping['label'] == 'admin':
            continue

        server = connection.execute(
            share_servers_table.select().where(
                alloc._mapping['share_server_id'] == (
                    share_servers_table.c.id))).first()

        # pylint: disable=no-value-for-parameter
        op.execute(network_allocation_table.update().where(
            alloc._mapping['id'] == network_allocation_table.c.id).values(
            {'share_network_subnet_id':
             server._mapping['share_network_subnet_id']}))

    # add a new column to share_servers.
    try:
        op.add_column(
            SHARE_SERVERS_TABLE,
            sa.Column('network_allocation_update_support', sa.Boolean,
                      nullable=False, server_default=sa.sql.false()))
    except Exception:
        LOG.error("Table %s could not add column "
                  "'network_allocation_update_support'.",
                  SHARE_SERVERS_TABLE)
        raise

    # drop subnet id foreign key from share servers.
    try:
        share_serves_fk_name = (
            "fk_share_servers_share_network_subnet_id_share_network_subnets")
        if connection.engine.name == 'mysql':
            op.drop_constraint(share_serves_fk_name, SHARE_SERVERS_TABLE,
                               type_="foreignkey")
        op.drop_column(SHARE_SERVERS_TABLE, 'share_network_subnet_id')
    except Exception:
        LOG.error("Table %s could not drop column 'share_network_subnet_id'.",
                  SHARE_SERVERS_TABLE)
        raise


def downgrade():
    """Remove share_server_share_network_subnet_mapping table and new columns.

    This method can lead to data loss because the share server can have
    more than one subnet.
    """
    try:
        share_serves_fk_name = (
            "fk_share_servers_share_network_subnet_id_share_network_subnets")
        op.add_column(
            SHARE_SERVERS_TABLE,
            sa.Column(
                'share_network_subnet_id', sa.String(36),
                sa.ForeignKey('share_network_subnets.id',
                              name=share_serves_fk_name),
            )
        )

        connection = op.get_bind()
        server_subnet_mappings_table = utils.load_table(
            SHARE_SERVER_SUBNET_MAP_TABLE, connection)
        share_servers_table = utils.load_table(SHARE_SERVERS_TABLE,
                                               connection)

        with sa.orm.Session(bind=op.get_bind()) as session:
            for server in connection.execute(share_servers_table.select()):
                subnets = session.query(
                    server_subnet_mappings_table).filter(
                        server._mapping['id'] == (
                            server_subnet_mappings_table.c.share_server_id)
                        ).all()

                if server._mapping['deleted'] != 'False' and len(subnets) > 1:
                    LOG.warning('Share server %s is not deleted and it '
                                'has more than one subnet (%s subnets), '
                                'the downgrade may cause an inconsistent '
                                'environment.',
                                server._mapping['id'], len(subnets))

                subnet_id = (
                    subnets[0].share_network_subnet_id if subnets else None
                )

                # pylint: disable=no-value-for-parameter
                op.execute(share_servers_table.update().where(
                    server._mapping['id'] == share_servers_table.c.id).values(
                    {'share_network_subnet_id': subnet_id}))

    except Exception:
        LOG.error("'share_network_subnet_id' field in the %s table could not "
                  "be created and populated from %s table.",
                  SHARE_SERVERS_TABLE, SHARE_SERVER_SUBNET_MAP_TABLE)
        raise

    try:
        op.drop_table(SHARE_SERVER_SUBNET_MAP_TABLE)
    except Exception:
        LOG.error("Failed to drop table %s.", SHARE_SERVER_SUBNET_MAP_TABLE)
        raise

    try:
        op.drop_column(SHARE_SERVERS_TABLE,
                       'network_allocation_update_support')
    except Exception:
        LOG.error("Table %s failed to drop the column "
                  "'network_allocation_update_support'.", SHARE_SERVERS_TABLE)
        raise

    try:
        network_allocation_fk_name = (
            "fk_network_allocation_subnet_id_share_network_subnets")
        if connection.engine.name == 'mysql':
            op.drop_constraint(network_allocation_fk_name,
                               NETWORK_ALLOCATIONS_TABLE,
                               type_="foreignkey")
        op.drop_column(NETWORK_ALLOCATIONS_TABLE, 'share_network_subnet_id')
    except Exception:
        LOG.error("Column 'network_allocations.share_network_subnet_id' from "
                  "table %s failed to drop.", NETWORK_ALLOCATIONS_TABLE)
        raise
